#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from scipy import integrate
import numpy as np
from palmiche.utils import xvg, tools
import matplotlib.pyplot as plt


def main(BoundRegion, radium, sigma, absolute_temperature, PmfXvgPath, integrator = "simpson"):
    """Get the IC50 from a umbrella simulation.
    The method is based in impose equal probabilities to be bound and unbound states.
    For that it is necessary the definition of the Bound region (which is done with BoundRegion)
    In the general case we have PMF = f(x) and we have the following interval:
    x0-----x1-----x2-----x3-----L----->inffinite
    Where x0 is the ligand inside the protein and inffinite is the ligand in the bulk.
    If we deffine the bound region in the interval [x1;x2]. Then the probability to be bound will be;
    Pbound = P_x1_x2 = I_x1_x2(exp(-E(x)/KbT)dx)/Z
    Where Z is the partion function of the system
    With the same analysis we could get the probability to be in all the rest of the {X}, that is suppouse that the system is unbounded. Therefore:
    Punbound = P_x0_x1 + P_x2_x3 + P_x3_xL
    where:
    P_x0_x1 = I_x0_x1(exp(-E(x)/KbT)dx)/Z
    P_x2_x3 = I_x2_x3(exp(-E(x)/KbT)dx)/Z
    P_x3_L = I_x3_L(exp(-E(x)/KbT)dx)/Z

    Generally in around x3 the PMF shows a flatt region, so we assume that beyond x3 the energy of the system will remain constant and equal to E(x3).
    For numerical reason we set thi value as the average of the last five points.
    if that is so:
    P_x3_L = I_x3_L(exp(-E(x)/KbT)dx)/Z = I_x3_L(exp(-E(x3)/KbT)dx)/Z = exp(-E(x3))(L-x3)/Z
    Finally imposed the equality:
    Pbound = Punbound
    P_x1_x2 = P_x0_x1 + P_x2_x3 + P_x3_L
    I_x1_x2(exp(-E(x)/KbT)dx)/Z = I_x0_x1(exp(-E(x)/KbT)dx)/Z + I_x2_x3(exp(-E(x)/KbT)dx)/Z + exp(-E(x3))(L-x3)/Z
    I_x1_x2(exp(-E(x)/KbT)dx) = I_x0_x1(exp(-E(x)/KbT)dx) + I_x2_x3(exp(-E(x)/KbT)dx) + exp(-E(x3))(L-x3)
    exp(-E(x3))(L-x3) = I_x1_x2(exp(-E(x)/KbT)dx) - I_x0_x1(exp(-E(x)/KbT)dx) - I_x2_x3(exp(-E(x)/KbT)dx)

    L = [I_x1_x2(exp(-E(x)/KbT)dx) - I_x0_x1(exp(-E(x)/KbT)dx) - I_x2_x3(exp(-E(x)/KbT)dx)] / exp(-E(x3)) + x3

    Finally, to get the IC50 because the cylinder restrain used on the umbrella simulation:
    V = np.pi*(radium + 2*sigma)**2*L*1E-24
    1E-24 is to convert from nm to dm. This implies that all the lenght units must be in [nm], Energy in [kJ/mol] and temperature in [K].
    
    IC50 = 1 / (Nav*V) [mol/L]
    There are several possibilities that are handled on the code:
        It is not necessary to extrapolate the data. If not, only the simulation data will be used
        The minimum definition could be from the first point to some distance or in between.

    Args:
        BoundRegion (float, or list of two floats): The definition of the bounded region
        radium (float): The radium used for the flat bottomed potential in [nm]
        sigma (float): The sigma of the flat bottomed potential (you could get it using the function tools.get_sigma()) in [nm]
        absolute_temperature (float): The simulation temperature in [K]
        PmfXvgPath (string path-like): The path to the xvg file with the PMF data (reaction coordinate[nm], energy [kJ/mol])
        integrator (str, optional): Type of integrator to be used. Defaults to "simpson". Valid methods are: 'simpson', 'trapezoid', 'trapz'

    Raises:
        ValueError: [description]

    Returns:
        [type]: [description]
    """
    # Selection of the integration method
    if integrator == "simpson":
        integrator = integrate.simpson
    elif integrator == "trapezoid":
        integrator = integrate.trapezoid
    elif integrator == "trapz":
        integrator = np.trapz
    else:
        raise ValueError(f"{integrator} is not a valid integration method. Valid methods are: 'simpson', 'trapezoid', 'trapz'")

    Nav = 6.02214076E23  
    data = xvg.XVG(PmfXvgPath).data
    data = np.array(sorted(data,key=lambda x:x[0])) # I am not able to use, the sort method of a numpy array in this case
    xmax = np.max(data[:,0])

    try:
        x1 = BoundRegion[0]
        x2 = BoundRegion[1]
    except:
        x1 = None
        x2 = BoundRegion

    if x1:
        data_x0_x1 = data[(data[:,0] <= x1)]
        data_x1_x2 = data[(x1 < data[:,0]) and (data[:,0] <= x2)]
        data_x2_x3 = data[(data[:,0] > x2)]
        raise ValueError('Not coded yet')
        # Esto no lo he porgramado todavia
        # Ahora tengo dos intervlos el minimo esta en el medio
    else:
        # The minimum is deffined from the first point of the data to X2           
        data_x0_x2 = data[(data[:,0] <= x2)]
        data_x2_x3 = data[(data[:,0] > x2)]


        # This is just to trnaslate the minimum and see the impact
        #to_add = np.array(len(data_x0_x2)*[(0,30)])
        #data_x0_x2 += to_add
        new_data = np.concatenate((data_x0_x2, data_x2_x3), axis = 0)
        fig, ax = plt.subplots(2,1,figsize = (16,9))
        ax[0].plot(new_data[:,0],new_data[:,1])
        ax[1].plot(new_data[:,0],tools.BoltzmannFactor(new_data[:,1], absolute_temperature))
        fig.savefig("pmf_boltzman_factor.pdf")

        # Take the last five energies to average, and get the Boltzmann Factor of the last point
        BoltzmannFactor_x3 = tools.BoltzmannFactor(np.mean(data[-5:,1]),absolute_temperature)

        P_x0_x2 = integrator(tools.BoltzmannFactor(data_x0_x2[:,1],absolute_temperature),data_x0_x2[:,0])
        P_x2_x3 = integrator(tools.BoltzmannFactor(data_x2_x3[:,1],absolute_temperature),data_x2_x3[:,0])
        #El problema aqui es que tengo que jugar con la data original, y cuando se me acaben los puntos entonces empezar a interpolar con un dX y empezar a adicionar puntos a la data, hasta que converja la tolerancia o hasta que se me acaben los puntos
        #Probablemente sea mejor correr sobre la data, probar una vez si con el maximo se obtiene convergencia (el signo sera imporrtante) sino ya ir directo a la suma de puntos 
        if P_x2_x3 > P_x0_x2:
            L = data_x2_x3[1,0]
            error = np.abs(integrator(tools.BoltzmannFactor(data_x2_x3[:1,1],absolute_temperature),data_x2_x3[:1,0]) - P_x0_x2)
            for i in range(2, len(data_x2_x3)):
                tmp_error = np.abs(integrator(tools.BoltzmannFactor(data_x2_x3[:i,1],absolute_temperature),data_x2_x3[:i,0]) - P_x0_x2)
                if tmp_error < error:
                    error = tmp_error
                    L = data_x2_x3[i,0]

        else:
            # n = N/Nav
            # C = n/V = N/(Nav*V)
            L = xmax + (P_x0_x2 - P_x2_x3) / BoltzmannFactor_x3

        V = np.pi*(radium + 2*sigma)**2*L*1E-24
        IC50 = 1 / (Nav*V)
        with open('ic50.txt', "w") as f:
            f.write(str(IC50))   
    return IC50




           

if __name__ == '__main__':
    import os, glob
    umbrellas = sorted(glob.glob('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_*'))
    for umbrella in umbrellas:
        try:
            ic50 = main(2, 0.4, 0.071, 303.15, os.path.join(umbrella, '7e27/sys_MMV007839_Cell_891_SP_param/windows/coord0_selected.xvg'))
            print(f"{os.path.basename(umbrella):>50}: IC50 = {ic50}")

        except:


            pass
    # with open('ic50.txt', "w") as f:
    #     f.write(str(ic50))   



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from palmiche.utils import tools
from glob import glob
def wham(select_file_names = "coord[0-9]_selected.dat", OMP_NUM_THREADS = 12):
    select_file_names = sorted(glob(select_file_names))
    #wham_general = ["gmx wham -it tpr_files.dat -if pullf_files.dat -o all_coords_selected -hist hist_all_coords_selected -unit kJ"]
    wham_individual_coords = [f"export OMP_NUM_THREADS={OMP_NUM_THREADS}; gmx wham -ac -temp 303.15 -zprof0 4.5 -bins 300 -unit kJ -nBootstrap 200 -bs-method hist -is {s} -ix pullx_files.dat -it tpr_files.dat -o {s.split('.')[0]} -hist hist_{s.split('.')[0]} -oiact iact_{s.split('.')[0]}.xvg -bsres bsResult_{s.split('.')[0]}.xvg -bsprof bsProfs_{s.split('.')[0]}.xvg" for s in select_file_names]
    for cmd in wham_individual_coords: #wham_general + 
        tools.run(cmd)

if __name__ == '__main__':
    wham()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from palmiche.utils import tools, pdb
import os, tempfile, argparse




'''
For pdb2gmx by default:
     LYS is protonated
     HIS, ASP and GLU deprotonated.
In case of:
    pKa > pH [HA] > [A-], it is protonated
    pKa < pH [HA] < [A-], it is deprotonated
The script will check this from the propka calculation and if the default option 
need to be changed, this will do automatically executing pdb2gmx with the command at the end:
This use propka3 not propka as command line.
You can specifie how you call propka on your computer with the flag --pka\n


From pdb2gmx

Which LYSINE type do you want for residue 14
0. Not protonated (charge 0) (LYN)
1. Protonated (charge +1) (LYS)

Which GLUTAMIC ACID type do you want for residue 3
0. Not protonated (charge -1) (GLU)
1. Protonated (charge 0) (GLH)

Which ASPARTIC ACID type do you want for residue 84
0. Not protonated (charge -1) (ASP)
1. Protonated (charge 0) (ASH)

Which HISTIDINE type do you want for residue 18
0. H on ND1 only (HID)
1. H on NE2 only (HIE)
2. H on ND1 and NE2 (HIP)
3. Coupled to Heme (HIS1)


Nomenclature of the proton in the pdb after pdb2gmx
HD1 ----> proton on delta (0 option)
HE2 ----> proton on epsilon (1 option) (the most away proton from the R chain)
could be both protons at the same time.
'''
 
def check_protonation(aa, pKa, pH):
    '''
    
    Parameters
    ----------
    aa : String
        The aa code, only ASP, GLU, HIS, LYS, case sensitive.
    pKa : float
        DESCRIPTION.
    pH : TYPE
        DESCRIPTION.

    Returns
    -------
    protonation_state : int
        0 or 1 for deprotonated or protoneted (1 and 2 for HIS).
    check : Bool
        True if the protonation match with the default of pdb2gmx.

    '''
    # 0 and 1 match with the input for pdb2gmx, except for HIS
    # The protonated state is 2 and the deprotonated state that we will use is
    # 1. H on NE2 only (HIE)
    # The next dict is only for check the protonation state. In HIS the default is deprotoneted but not precisilly 0. 
    # When deprotonated, gmx will try to figure out which state is presented (delta or epsilon)
    # But for convenience, we could use 0 and one to check if it is protoneated or deprotoneted and 
    # at the end sum 1. That means that this will not give nor the delta protonation neither Coupled to Heme (HIS1) states of HIS only
    # epsilon and protonated. This could be improved in the future and get more information from propka. For our propouse it is sufficient.
    # for the rest of the aa.
    
    gmx_default = {
        'ASP': 0,    #deprotonated
        'GLU': 0,    #deprotonated
        'HIS': 0,    #deprotonated
        'LYS': 1,    #protonated
        }
    
        
    protonation_state = 0    
    check = True
    
    if pKa > pH:
        protonation_state = 1
    if gmx_default[aa] != protonation_state:
        check = False
    if aa == 'HIS': protonation_state += 1 #(1---->H on NE2 only (HIE), 2----> dobleprotonated )
    
    return protonation_state, check



def pka2gmx(pdb_path,
            protein_forcefield = "amber99sb-ildn",
            water_forcefield = "tip3p",
            ignh = True,
            ph = 7.0,
            pKa = "propka3",
            addph = [],
            addresifile = [],
            output = "system_GMX.pdb",
            pKa_file_out = True,
            itp_top_out = True,
            num_file = True):
    # Getting cwd and changing to temporal and storaging names
    cwd = os.getcwd()
    tmpdir = tempfile.TemporaryDirectory()
    pdb_path = os.path.abspath(pdb_path)
    output = os.path.abspath(output)
    file_name = f"{os.path.basename(pdb_path).split('.')[0]}"
    pKa_file = f"{file_name}.pka"

    # Totally inefficient but needed in case of force fields in the directory
    for item in tools.list_if_dir(cwd):
        if item.endswith('.ff'):
            # I need to copy everything and not just the force field because some times you need another files outside of the ff
            # The only directories that I will copy are the ones with ends in .ff
            
            [tools.cp(file, tmpdir.name, r = True) for file in tools.list_if_file()]
            [tools.cp(dir, tmpdir.name, r = True) for dir in tools.list_if_dir() if dir.endswith('.ff')]
            break
    os.chdir(tmpdir.name)

 
    # This part if for multiple pH on the protein. Ask for the pH to consider and the files where the resi number are.

    aa_ph_manual_input = {}
    if addph and len(addph) == len(addresifile): # If you specified more than one pH and each pH has one and only resifile

        print(f'The main ph (1) is {ph}')    
        for (ph_i, resifile_i) in zip(addph,addresifile):

            with open(resifile_i, 'r') as f:
                resifile_i_lines = f.readlines()
            # Here I am assigning to each aa the corresponded specified ph. I am inverted the input, the normal input is the ph and a file with the aa that we want at that pH
            # Now I change to all the aa in the file I assign the pH, if the next file with the next pH has the same aa the pH value will be update because aa_ph_manual_input is a dictionary. 
            for resi in resifile_i_lines:
                aa_ph_manual_input[resi] = float(ph_i)

    if os.path.exists(pKa_file):
        print(f"Propka was not call because the file {pKa_file} exist.")
    else:
        tools.run(f"{pKa} {pdb_path}")


    with open(pKa_file, 'r') as f:
        pKa_lines = f.readlines()



    togmx = {
            'ASP': [],
            'GLU': [],
            'HIS': [],
            'LYS': [],
            }
    """
    In this section we are looking for the propka prediction and constructing 
    togmx: a dict of dicts that contains all the information
    {"resi_id":resi_id,"chain":chain,"prot_state":protonation_state,"check":check}
    """
    for i in range(len(pKa_lines)):
        if 'SUMMARY OF THIS PREDICTION' in pKa_lines[i]:
            j = i + 2
            while('---------') not in pKa_lines[j]:
                aa = pKa_lines[j][3:6]
                resi_id = int(pKa_lines[j][6:10])
                chain = pKa_lines[j][11]
                pKa = float(pKa_lines[j][13:21])

                if aa in togmx:
                    if resi_id in aa_ph_manual_input: # If the aa was specified with some additional pH, then check with this one.
                        protonation_state, check = check_protonation(aa, pKa, aa_ph_manual_input[resi_id])
                        print(f"The resi {'-'.join([aa, str(resi_id), chain])} was tested with the pH {aa_ph_manual_input[resi_id]}\n")
                    else:
                        protonation_state, check = check_protonation(aa, pKa, ph)
                    togmx[aa].append({"resi_id":resi_id, "chain":chain, 
                                      "prot_state":protonation_state, "check":check})

                j += 1
            break

    '''
    Propka order the aa by chain and by resi number (the same as pdb2gmx)
    
    Only we need to be sure to ordered the sequence of comands to pdb2gmx
    
    The order that pdb2gmx ask you is by alphabetic order in the case that more than one
    flag was needed. For example, if on GLU need to be fixed and two ASP
    pdb2gmx will ask you first for ASP (the chain A, all the residues of chain A
    ordered , then chain B, all residues ordered respect to the residue number). 
    Then it will ask you for GLU in the same manner
    
    Because pdb2gmx use some optimization of the H+-bond network. In order to take
    advantages of this. We check which protonation found pdb2gmx and used.
    Only if HIS give False in the check
    '''
    tools.run(f"gmx pdb2gmx -f {pdb_path} -o tmp.pdb -ff {protein_forcefield} -water {water_forcefield} -ignh -merge all")
    [tools.rm(item) for item in ["*.top","*.itp"]]
    #tools.run('grep HIS tmp.pdb | egrep "HD1|HE2" > tmp.pdb.tmp; mv tmp.pdb.tmp tmp.pdb')
    #Here I am getting all the HISs fo my system
    # This is force field depended, because the HIS could have other names in different force fields, the function fix_pdb of assemble_MD could help for it
    # This will only work for amber, but could, of course be extended.
    protonation_HIS = []
    for resi in pdb.PDB('tmp.pdb').get_residues():
        if resi.resName in ["HIS", "HIE","HIP"]: #Any possible nomenclature of HIS in the AMBER force field except for HIS1
            resatoms = [atom.name for atom in resi.atoms]
            if not set(['HD1', 'HE2']) - set(resatoms): # This means that both atoms are in resatoms
                protonation_HIS.append(2)
            elif 'HE2' in resatoms:
                protonation_HIS.append(1)
            elif 'HD1' in resatoms:
                protonation_HIS.append(0)
            else:
                pass

    if len(protonation_HIS) == len(togmx['HIS']):
        for i in range(len(togmx['HIS'])):
            # This doble check means that we only consider the non-protoneted
            #HIS to catch if is epsilon or delta. But in the case of the 
            #prediction of protonation we will keep the Propka prediction.
            #If appear a case when pdb2gmx predict protonationi and Propka did
            #not. We will keep the epsilon deprotonation state.
            if togmx['HIS'][i] != 2 and protonation_HIS[i] != 2: 
                togmx['HIS'][i]["prot_state"] = protonation_HIS[i] # only use the information if the proton is delta or epsilon
    else:
        print('''WARNING: Something terrible wrong hapened with the pdb2gmx comprobation of HIS.
              Therefore the deprotonated HIS will be considered as the epsilon only.
              ''')
    tools.rm('tmp.pdb')

    file = open('num.txt', 'wt')
          
    print('The field num.txt have the protonation state for\n')
    aa_flag_cmd = ""

    # This is for the construction of the comand and to know which aa need to be change
    aa2change = []
    ID_chains = set()
    resi_number = set()
    
    for aa in sorted(togmx): # in this way we sorted the keys and aa2change will append the aa in a sorted way
        check = set()
        for resi in togmx[aa]:
            ID_chains.add(resi["chain"])
            resi_number.add(resi["resi_id"])
            check.add(resi["check"])
            
        if False in check:
            aa_flag_cmd += f"-{aa.lower()} "
            aa2change.append(aa)
    
    ID_chains = sorted(ID_chains) #When we sorted a set, automatically you get a list
    resi_number = sorted(resi_number)
    
    # It is important the order of num.txt. for gmx (chain alphabetic--> aa alphabetic --> ID of each aa)

    for id_c in ID_chains: #(A, B, C, ...)
        for aa in aa2change: #(ASP, GLU, HIS, LYS), only for those aa that we cheked that need change before
            for number in resi_number:# (1, 2, 3, 4, ...)
                for resi in togmx[aa]: # Loops for each residue of each aa (ASP, GLU, HIS, LYS)
                    if resi["chain"] == id_c and resi["resi_id"] == number: #Here we are matching the residue with Chain id_c, aminoacid name aa, and residue number resi_id
                        print(f"{aa}-{resi}-{id_c} \t\t {resi['prot_state']}")
                        file.write(f"{resi['prot_state']}\n")
                        
    file.close() # It is important to close the file before cat it. 
    
    
    cmd = f"export GMX_MAXBACKUP=-1; cat num.txt | gmx pdb2gmx -f {pdb_path} -o {output} -ff {protein_forcefield} -water {water_forcefield} {aa_flag_cmd}"
    if ignh: cmd += " -ignh" #Just for strange cases that you are not able to remove the H of you system
    tools.run(cmd)
    print('The command used:')
    print(cmd)

    os.chdir(cwd)

    # Coping from temporal directory
    if pKa_file_out:
        tools.cp(os.path.join(tmpdir.name, pKa_file), '.')
    if itp_top_out:
        [tools.cp(item, '.') for item in [os.path.join(tmpdir.name, "*.top"), os.path.join(tmpdir.name, "*.itp")]]
    if num_file:
        tools.cp(os.path.join(tmpdir.name, 'num.txt'), '.')
if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='pKa2pdb2gmx.')
    parser.add_argument("pdb_path", 
                        help = "The path to the pdb file to process.",
                        type = str)
    parser.add_argument("-pf", 
                        dest="protein_forcefield",
                        default = "amber99sb-ildn",
                        help = "The protein force field to use with GROMACS",
                        type = str)
    parser.add_argument("-wf", 
                        dest="water_forcefield",
                        default = "tip3p",
                        help = "The water force field to use with GROMACS",
                        type = str)   
    parser.add_argument("-ph", 
                        dest="ph",
                        default = 7.0,
                        help = "The general value of pH.",
                        type = float)
    parser.add_argument("-pKa", 
                        dest="pKa",
                        default = "propka3",
                        help = "The propka executable name.",
                        type = str)
    parser.add_argument("-ignh", 
                        dest="ignh",
                        nargs="?",
                        const=True,
                        default=False,
                        type=bool)
    parser.add_argument("-addph", 
                        dest="addph",
                        nargs = "+",
                        default = [],
                        help = "The other pH to condsider for each value must be file with aa to take into account with the flag addresifile.",
                        type = float)  
    parser.add_argument("-addresifile", 
                        dest="addresifile",
                        nargs = "+",
                        default = [],
                        help = "The other pH to condsider for each value must be file with aa to take into account with the flag addresifile.",
                        type = str)
    parser.add_argument('-output',
                        help='The output path, Default is system_GMX.pdb',
                        dest='output',
                        default = "system_GMX.pdb",
                        type = str) 
    parser.add_argument('-pKa_file_out',
                        help='If you want to output the pKa file. Default is False',
                        nargs="?",
                        dest='pKa_file_out',
                        const=True,
                        default=False,
                        type=bool)                   
    parser.add_argument('-itp_top_out',
                        help='If you want to output the itp and top files. Default is False',
                        nargs="?",
                        dest='itp_top_out',
                        const=True,
                        default=False,
                        type=bool) 
    parser.add_argument('-num_file',
                        help='If you want to output the file num file. This file contains the info of the protonation used for pdb2gmx. Default is False',
                        nargs="?",
                        dest='num_file',
                        const=True,
                        default=False,
                        type=bool)

    args = parser.parse_args()
    pka2gmx(args.pdb_path,
            protein_forcefield = args.protein_forcefield,
            water_forcefield = args.water_forcefield,
            ph = args.ph,
            pKa = args.pKa,
            ignh = args.ignh,
            addph = args.addph,
            addresifile = args.addresifile,
            output = args.output,
            pKa_file_out = args.pKa_file_out,
            itp_top_out = args.itp_top_out,
            num_file = args.num_file)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import palmiche.utils.pdb as pdb
from palmiche.utils.assembleMD import fix_pdb
import palmiche.utils.tools as tools
import os.path

    


def databox(input_file, exp_structure, pymol_path_cmd = "pymol", lig_name = None, distance = 0.4):
    
    if not os.path.isfile(input_file):
        raise FileNotFoundError(f"The file {input_file} doesn't exist or is not accesible")
    if not os.path.isfile(exp_structure):
        raise FileNotFoundError(f"The file {exp_structure} doesn't exist or is not accesible")
        
    """   
    This function (fix_pdb) will export the pdb file:
        1-PROTEIN.pdb
    The next steps on this function will make use of PROTEIN.pdb
    
    """
    PDB = fix_pdb(input_file)
    tools.run(f"{pymol_path_cmd} PROTEIN.pdb {exp_structure} &")
    answer1 = "a"
    while answer1 != "c":
        answer1 = input("""You need to align the pdb structure of the PDB file that
          has the ligands to the structure for fix_pdb(input_file), 
          and create and save a 
          pdb file under the name PROTEIN.pdb (the same name as the 
                exported structure) that has the protein and the ligands 
          aligned. This could be done in Pymol.
          0-Open input_file and the experimental pdb. In this order, so, the ligands
           will be at the end of the file. (this should had done 
            automathically)
          1-Align the experimental or the one that has the ligands orientation
          to input_file, center the view in PROTEIN to seee the alignment.
          2-Slect all the secuence of input_file
          3-Select also the ligands
          4-Export the selection as input_file (Replacing the existing PDB file)\n\n
          
          If you already finished this step type 'c' to continue: """).lower()
    
    """
    Rectified again the PDB is needed becasue Pymol change the position of the
    TER flag.
    """
    
    PDB = pdb.PDB("PROTEIN.pdb")#I nead to load the new PDB from the pymol work
    PDB.fix_chains()
    
    #Now the name of the ligands will be changed    
    if lig_name:
        for atom in PDB.atoms:
            if atom.resName == lig_name:
                atom.resName = f"LI{atom.chainID}"
    #write the new PDB file with the Ligan names corrected
    PDB.write()
    chains = PDB.get_chains()
    #Creating the index file and  the pdb to get the box dimensions
    #The first step is the generations of the chain index with make_ndx
    with open("opt", "w") as opt:
        for chain in chains:
            opt.write(f"chain {chain}\n")
        opt.write("q\n")
    tools.run("cat opt | gmx make_ndx -f PROTEIN.pdb -o index")
    
    with open("opt", "w") as opt:
        for chain in chains:
            opt.write(f"\"SELECT_{chain}\" group \"ch{chain}\" and same residue as within {distance} of resname LI{chain};\n")
    tools.run("gmx select -s PROTEIN.pdb -sf opt -n index.ndx -on index.ndx")
    
    #deleting the line _f0_t0.000 in the file
    with open("index.ndx", "r") as index:
        data = index.read()
        data = data.replace("_f0_t0.000","")
    with open("index.ndx", "w") as index:
        index.write(data)
    
    
    #create the PDB selection, get the center of the box and edges, delete the pdb file
    boxes = {}
    for chain in chains:
        tools.run(f"echo \"SELECT_{chain}\" | gmx editconf -f PROTEIN.pdb -o SELECT_{chain}.pdb -n index.ndx")
        SELECT = pdb.PDB(f"SELECT_{chain}.pdb")
        boxes[chain] = {"boxcenter": SELECT.box_center(), "boxsize": SELECT.edges()}
        tools.rm(f"SELECT_{chain}.pdb")
    [tools.rm(item) for item in ["opt", "index*", "#*"]]
    
    return boxes
    
if __name__ == '__main__':
    boxes = databox("step5_input.gro", "7e27.pdb", pymol_path_cmd = "/home/ale/pymol/pymol", lig_name = "HV6")
    print(boxes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__description__ = \
"""
Take a file containing frame number vs. distance, then identify frames that 
sample those distances at approximately window_width.  Optionally takes an
arbitrary number of template files.  The program searches the contents of these
files for a search string (by default XXX), replaces the search string with the
frame number, then writes out each file with a unique, frame-specific filename.
This last feature means that this script can be used to automatically generate
input files to run each umbrella sampling simulation. 

v. 0.1: In this version I (Alejandro) made modifications in the way to select
the umbrella windows (below explained why) and also some other small changes respect
to the first version of Michael J. Harms [harmsm@gmail.com].
"""
__usage__ = "setupUmbrella.py --disfile(-d) distance_file --interval1(-i1) main_window_width [--interval2(-i2) secondary_window_width --number(-n2) number_of_secondary_interval --template(-t) template_file]"

import os, numpy as np
from argparse import ArgumentParser
 
def readDistanceFile(distance_file):
    with open(distance_file, 'r') as f:
        lines = f.readlines()
    return [(int(line.split()[0]), float(line.split()[1])) for line in lines]

def sampleDistances(table_distance, window_width, first = False, **keywords):
    """
    Go through the distances list and sample frames at window_width.  
    Appropriate samples are identified by looking forward through the
    distances to find the one that is closest to current_distance +
    window_width.  


    Args:
        table_distance (list): This list is returned by the function readDistanceFile()
        window_width (float): This is the width of each window.py
        first (bool, optional): In case that the first frame want to be for sure included. Defaults to False.
        **keywords (optional): minimum and maximum. 
            In case you want to specified the range for the umbrella, if not is specified minimum and/or maximum. The real
            extreme value(s) of table_distance will be used.
            In the case of maximum, if you want to pick the value you need to provide, for example:
                maximum + window_width, because if not the np.arange will not take into account the last number.py. If you don't specify
                the maximum value the function will add automatically window_width to maximum.

    Returns:
        [list]: [List of window indexes]
    """
    
    distance = [d[1] for d in table_distance]
    #Here we are taken the interval to be sampled
    try:
        minimum = keywords["minimum"]
    except:
        minimum = min(distance)
    try:
        maximum = keywords["maximum"]
    except:
        maximum = max(distance) + window_width


    windows = []
    windows_index = []
    miss_center = []
    for center in np.arange(minimum, maximum, window_width):
        onward = [abs(com-center) for com in distance]
        min_index = onward.index(min(onward))
        if center - window_width/2 <= distance[min_index]\
        and distance[min_index] <= center + window_width/2\
        and distance[min_index] not in windows:
            #print(center, distance[min_index])
            windows.append(distance[min_index])
            windows_index.append(min_index)
        else:
            miss_center.append(center)
    
           
    if miss_center:
        print(f"Warning: For the interval {window_width} the following center of windows were not found: {miss_center}")
    if first:
        if 0 not in windows_index:
            windows_index.insert(0,0)
    return windows_index

def createOutputFile(template_file,frame_number,search_string="XXX"):
    """
    Look for instances of the search string in a template file and replace with
    the frame number in a new file.
    """

    out_file = f"frame-{frame_number}_{template_file}"

    # Prompt the user before wiping out an existing file
    if os.path.exists(out_file):
        answer = input(f"{out_file} exists!  Overwrite (y|n)?")
        answer = answer[0].lower()
        if answer != "y":
            return None
    
    # Read the contents of the template file
    with open(template_file,'r') as f:
        file_contents = f.read()

    # Write out the template file contents, replacing all instances of 
    # the search string with the frame number
    with open(out_file,'w') as f:
        f.write(file_contents.replace(search_string,str(frame_number)))

def main(disfile, window_width, first = False, template = None, txt_summary = 'conf_windows_summary.txt',  **keywords):
    # Figure out which frames to use
    distance_table = readDistanceFile(disfile)
    #Because first = True, the first conformation is always added
    sampled_indexes = sampleDistances(distance_table,float(window_width),first=first, **keywords)

    # If any template files were specified, use them to make frame-specific
    # output
    if template != None:
        print(f"Creating frame-specific output for {template}:")
        for i in sampled_indexes:
            frame = distance_table[i][0]
            createOutputFile(template,frame,search_string="XXX")    
    
    # Print out summary of the frames we identified 
    to_print = sorted([distance_table[index] for index in sampled_indexes], key=lambda x:x[1])

    
    out_string = f"{'window_ID':<15}{'frame':<10}{'dist':<10}{'delta_dist':<10}\n"
    windows = {}
    for i, item in enumerate(to_print):
        window_ID = i+1
        frame = item[0]
        dist = item[1]
        if i == 0:
            delta_dist = f"{'NaN':<10}" 
        else:
            prev_dist = to_print[i-1][1]
            delta_dist = f"{(dist - prev_dist):<10.3}"
        
        out_string += f"{window_ID:<15}{frame:<10}{dist:<10.3}{delta_dist}\n"
        windows[window_ID] = {'frame': frame, 'dist': dist, 'delta_dist': delta_dist.strip()}
    
    if txt_summary:
        with open(txt_summary, 'w') as o:
            o.write(out_string)
    return windows

     


def mainxxxx(disfile, interval1, interval2 = 0, num_windows_interval2 = 1, template = None):
    """
    This is only the main function, the description is forward. type:
        -h to get help

    """    
    
    # Figure out which frames to use
    distance_table = readDistanceFile(disfile)
    #Because first = True, the first conformation is always added
    sampled_indexes = sampleDistances(distance_table,float(interval1),first=True)
    
    #If we want to add extra windows near to the window with the lowest COM:
    #    interval2 controls the width and num_windows_interval2 how many.
    if interval2:
        sampled_indexes += sampleDistances(distance_table,float(interval2))[1:int(num_windows_interval2)+1]
        
    sampled_indexes = list(set(sampled_indexes)) #to remove the duplicate items
    
    # If any template files were specified, use them to make frame-specific
    # output
    if template != None:
        print("Creating frame-specific output for {template}:")
        for i in sampled_indexes:
            frame = distance_table[i][0]
            createOutputFile(template,frame,search_string="XXX")    
    
    # Print out summary of the frames we identified 
    to_print = sorted([distance_table[index] for index in sampled_indexes], key=lambda x:x[1])

    
    out_string = f"{'window_ID':<15}{'frame':<10}{'dist':<10}{'delta_dist':<10}\n"
    windows = {}
    for i, item in enumerate(to_print):
        window_ID = i+1
        frame = item[0]
        dist = item[1]
        if i == 0:
            delta_dist = f"{'NA':<10}" 
        else:
            prev_dist = to_print[i-1][1]
            delta_dist = f"{(dist - prev_dist):<10.3}"
        
        out_string += f"{window_ID:<15}{frame:<10}{dist:<10.3}{delta_dist}\n"
        windows[window_ID] = {'frame': frame, 'dist': dist, 'delta_dist': delta_dist.strip()}
    
    with open('resume_conf_windows.txt', 'w') as o:
        o.write(out_string)
    return windows

if __name__=='__main__':
    #Getting the arguments
    #setupUmbrella.py --disfile(-d) distance_file --interval(-i) window_width --template(-t) [template_file1 template_file2 ...]
    parser = ArgumentParser(description='Getting umbrella windows (This script assumes an increases in the COM)')
    parser.add_argument('-d', '--disfile', dest='disfile',
                    help="""The name of distance file in the format n*2, where
                    the two columns are the the integer identify of the 
                    conformer and the center of mass""")
    parser.add_argument('-i1', '--interval1', dest='interval1',
                    help="""
                    The main interval distance between each windows. In nm 
                    """)
    parser.add_argument('-i2', '--interval2', dest='interval2', default = 0,
                    help="""
                    The default is 0,  not usedT.his interval is used in case
                    do you want to add some other windows respect to the first
                    window (the one with lower COM, which is not mandatory the
                    starting configuration for the pulling, could be another one.
                    For example, if you want, close to the first window, one 
                    window of 0.1 and your main interval is 0.2 (controlled by interval1).
                    Then you need to set --interval2(-i2) 0.1. If you want more than one
                    windows close to the first one you need to change 
                    num_windows_interval2, that by default is 1. If the new windows is 
                    already included with the main interval, then it will be 
                    dismissed.
                    """)
    parser.add_argument('-n2', '--num_windows_interval2', dest='num_windows_interval2', default=1,
                    help="""
                    Control of how many windows are created with the 
                    secondary interval close to the window of lower COM. 
                    Default is 1.
                    """)
    parser.add_argument('-t', '--template', dest='template',
                    help="""
                    The template for the execution of GROMACS (a shell script)
                    """)
    args = parser.parse_args()
    
    for item in args.__dict__:
        if item not in ['interval2','template']:
            if args.__dict__[item] == None:
                raise SyntaxError(f"The argument \"{item}\" is missing. The right way to call the script is (in \"[]\" the optional arguments): {__usage__}")
    
    main(args.disfile, args.interval1, interval2 = args.interval2, num_windows_interval2 = args.num_windows_interval2, template = args.template)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import tools
def wham():
    ligands = [0,1,2,3,4]
    number_windows = 87
    number_coords = 3*len(ligands)#THis case I use 3 because I use 2 pull restrain peer coord. orientation and loss of channel entrance 
    select_file_names = []
    for i in range(len(ligands) + 1):
        if i < len(ligands):
            row = tools.zerolistmaker(number_coords)
            row[i] = 1
            select_file_names.append(f"coord{i+1}_selected.dat")
            with open(f"coord{i+1}_selected.dat", "w") as select:    
                for j in range(number_windows):
                    select.write(f"{' '.join([str(r) for r in row])}\n")
        else:
            row = tools.zerolistmaker(number_coords)
            row[:len(ligands)] = len(ligands)*[1]
            select_file_names.append(f"coord0_selected.dat")
            with open(f"coord0_selected.dat", "w") as select:    
                for j in range(number_windows):
                    select.write(f"{' '.join([str(r) for r in row])}\n")

                               
         

    wham_individual_coords = [f"gmx wham -ac -temp 303.15 -zprof0 4.5 -bins 300 -unit kJ -nBootstrap 200 -bs-method hist -is {s} -ix pullx_files.dat -it tpr_files.dat -o {s.split('.')[0]} -hist hist_{s.split('.')[0]} -oiact iact_{s.split('.')[0]}.xvg -bsres bsResult_{s.split('.')[0]}.xvg -bsprof bsProfs_{s.split('.')[0]}.xvg" for s in select_file_names]

    for cmd in wham_individual_coords:#wham_general + wham_individual_coords:
        tools.run(cmd)
if __name__ == '__main__':
    wham()


#!/usr/bin/env python
'''
Author(s): Jeremy Lapierre <jeremy.lapierre@uni-saarland.de>, 
\nDescription: This script modifies gromacs topology files to simulate with heavy hydrogens. This
allows to decrease angular and out-of-plane motions involving hydrogens and therefore allowing to
increase the time step. To conserve the total mass, the heavy atoms connected to hydrogens will
also have a proportionally reduced mass. Example for methyl group with factor of 3 (default, see
argument descriptions at the end of the help) :\n
           H (1.008)                     H (3.024)
           |                             |
(1.008) H--C--H (1.008)  ==>  (3.024) H--C--H (3.024)
        (12.01)                       (5.9620)

(FLOAT): Atom mass\n

Another approach is to use virtual sites, however it is not possible to do the update on the GPUs
with virtual sites. Therefore, using heavy hydrogens allow to double the time step AND to use
gmx mdrun -update gpu and gain further significant speed-up.
It is atm only possible to use a factor of 4 with pdb2gmx -heavyh thus the need of this script.

If you are simulating proteins, you can try to constrain all bonds: "constraints = all-bonds" in
the .mdp file. But if the system is too big you won't be able to do the updates on the gpus.\n

Some script launch examples:\n
1. heavyh.py : a .top file is looked in current dir and all .itp files in current directory
and included in the .top will be processed as well. Files in the *posre*.itp form and itp files
containing "/" will be ignored. This has been done like this to avoid processing files like
"amber99sb-ildn.ff/forcefield.itp", files in the root force field directory and more generally to
have better controlled of what is processed where. All hydrogen masses are multiply by default
factor (i.e. 3).
2. heavyh.py -p topol.top : same as 1. but -p needed if several .top in directory.
3. heavyh.py -p topol1.itp topol2.itp: only processes the .itp files given, does not look
for .top.
4. heavyh.py -p topol.top -factor 2.5 : same as 2. but all hydrogen masses are multiply by a
factor of 2.5.
5. heavyh.py -p topol_heavyH_3.top -revert 3.024 -factor 3 -o _revertHMR : revert HMR on hydrogens
having a mass of 3.024amu (have been originally increased by a factor 3)

Ref:
1. K Anton Feenstra, Berk Hess and Herman J C Berendsen. Improving Efficiency of Large Time-Scale
Molecular Dynamics Simulations of Hydrogen-Rich Systems. Journal of Computational Chemistry, 1999.
2. Chad W. Hopkins, Scott Le Grand, Ross C. Walker and Adrian E. Roitberg. Long-time-step molecular
dynamics through hydrogen mass repartitioning. Journal of Chemical Theory and Computation, 2015.
3. Curtis Balusek et al. Accelerating Membrane Simulations with Hydrogen Mass Repartitioning.
Journal of Chemical Theory and Computation, 2019.

TODO/Notes:\n1. It is not possible to have several Hydrogens with different masses in the same
topology file yet. If those topologies are in different .itp files included in the .top, this will
be all fine though. However, for reverting back HMR with -revert flag, make sure all your hydrogens
have the same mass as the one you've given to the flag. If this is not the case, process your
topologies separately.
2. If included .itp files in .top are not in the same directory as the .top, the script will not
process it.
3. This scripts assumes that [ atoms ] block is followed by the [ bonds ] block 
(not [ pairs ], [ dihedrals ], etc.), this has been the case in all the case systems so far, but if
this leads to issues, it should be "not too hard" to change this in the script.

'''

import argparse
from glob import glob
import re
from os import path
from shutil import move
import itertools


class NoHydroError(Exception):
    """Customized exception"""
class SeveralHydroMass(Exception):
    """Customized exception"""

def file_checker(top, itp, suff):
    """This function checks for .itp (not containing 'posre' or '/') included in .top and try to
    look for them in the current directory or in the directory name of the .top file given by -p.\n
    Check if .itp/.top files have an [ atoms ] block to modifiy, else remove it from input files.\n
    Finally, builds output names for the topologies"""


    if top != []:
        print(f'Working on following .top file: {top}\n')
        regxitp = re.compile(r'(?<=#include ")(?!.*(\b\/\b|posre)).*(?=")')
        with open(top[0], 'r') as intop:
            for line in intop:
                if regxitp.search(line):
                    if regxitp.search(line).group() not in itp:
                        itp.append(regxitp.search(line).group())

        #Check if those .itp are in current directory or in the dirname of .top given by -p
        for i in itp:
            if not path.isfile(i):
                dirname_of_top = path.dirname(top[0])
                print(f"{i} included in .top not found in './', "
                      f"let's look in dirname of the -p argument (i.e. {dirname_of_top})")
                guess_itp_path = f'{dirname_of_top}/{i}'
                if path.isfile(guess_itp_path):
                    print(f'{i} found in {dirname_of_top}')
                    itp[itp.index(i)] = guess_itp_path
                else:
                    print(f"{i} included in .top also not present in {dirname_of_top}:"
                          f" exiting script")
                    raise SystemExit()

    topols = top + itp

    regx_atmblock = re.compile(r'^\[ atoms \]\n')
    topol2rm = []

    for i in topols:
        keep = 0
        with open(i, 'r') as topfile:
            for line in topfile:
                if regx_atmblock.search(line):
                    keep = 1
        if keep == 0:
            print(f'No [ atoms ] block in {i}, this file will be ignored')
            topol2rm.append(i)

    topols = [topfile for topfile in topols if topfile not in topol2rm]

    print(f'\nThe following .itp and/or .top will be used for generating heavy H topologies:'
          f'\n{topols}\n')

    topolout = [re.sub(r'(\.itp|\.top)', '', path.basename(name))
                +suff
                +'.'
                +name.split('.')[-1] for name in topols]

    topolin_topolout = {topols[i]: topolout[i] for i in range(len(topols))}

    return topolin_topolout

def top_parser(topol, revert):
    """This function is parsing the topol file and creates 2 main dictionaries: toplines and
    atomblock. The file is read line by line such that everything that is not [ atoms ] is put
    in toplines associted to their line index. All [ atoms ] lines are put in atomblock with
    associated line index. Bondlist is the lists of bonds."""

    toplines = {}
    atomblock = {}
    hydrogen_mass = set()
    hydroid2line = {}
    heavyid2line = {}
    bondlist = []
    endblock = re.compile(r"^\[ *")

    #Sanity check variable
    massin = 0

    with open(topol, 'r') as top:
        for index, atom_line in enumerate(itertools.takewhile(lambda x: '[ atoms ]' not in x, top)):
            toplines[index] = atom_line

        toplines[max(toplines.keys())+1] = '[ atoms ]\n'

        for index, atom_line in enumerate(itertools.takewhile(lambda x: '[ bonds ]' not in x, top),
                                          max(toplines.keys())+1):
            #must be before split(), because index is out of range otherwise but NO ERROR MESSAGE!
            if atom_line == '\n':
                toplines[index] = atom_line
                continue
            atom_line_list = atom_line.split()
            #print(atom_line_list)
            if atom_line_list[0] == ";":
                toplines[index] = atom_line
            elif len(atom_line_list) >= 8:
                if revert != 0:
                    if float(atom_line_list[7]) == revert:
                        hydrogen_mass.add(atom_line_list[7])
                        hydroid2line[atom_line_list[0]] = index
                        atomblock[index] = atom_line_list
                        massin += float(atom_line_list[7])
                    else:
                        heavyid2line[atom_line_list[0]] = index
                        atomblock[index] = atom_line_list
                        massin += float(atom_line_list[7])
                else:
                    if 0.99 <= float(atom_line_list[7]) <= 1.1:
                        hydrogen_mass.add(atom_line_list[7])
                        hydroid2line[atom_line_list[0]] = index
                        atomblock[index] = atom_line_list
                        massin += float(atom_line_list[7])
                    else:
                        heavyid2line[atom_line_list[0]] = index
                        atomblock[index] = atom_line_list
                        massin += float(atom_line_list[7])

        #Sanity check
        if hydrogen_mass == set():
            raise NoHydroError
        if len(hydrogen_mass) > 1:
            error = (f'Hydrogens with different masses where found in {topol}, this is not'
                     f' supported yet. Please remove outputs and improve the script.')
            raise SystemExit(error)
        #End of sanity check

        toplines[max(toplines.keys())+1] = '[ bonds ]\n'

        for index, atom_line in enumerate(top, max(toplines.keys())+1):
            if endblock.search(atom_line):
                next_block = atom_line
                break
            if atom_line == '\n':
                toplines[index] = atom_line
                continue
            atom_line_list = atom_line.split()
            if atom_line_list[0] == ";":
                toplines[index] = atom_line
            else:
                bondlist.append(atom_line.split()[0:2])
                toplines[index] = atom_line

        toplines[max(toplines.keys())+1] = next_block

        for index, atom_line in enumerate(top, max(toplines.keys())+1):
            toplines[index] = atom_line

    hydrogen_mass = list(hydrogen_mass)

    return (toplines, atomblock, float(hydrogen_mass[0]),
            hydroid2line, heavyid2line, bondlist, massin)


def atom_block_modifier(hydrogen_mass, massfactor, bondlist, hydroid2line, heavyid2line, atomblock):
    """This function iterate over the bonds columns to check which atom ids are linked to an H.
       Heavy atoms are then retrieved in atomblock thanks to heavyid2line in order to decrease
       their masses. Hydrogens atoms are also retrieved in atomblock but thanks to hydroid2line and
       their masses are increased according to the massfactor. Usually H are in 2nd col of the
       [ bonds ] block, but the elif make sure to also check col 1 just in case. The use of set
       objects makes sure to do this really fast in contrast to lists (O(n) vs O(1))."""

    new_hydrogen_mass = hydrogen_mass*massfactor
    massdelta = hydrogen_mass - new_hydrogen_mass

    #Sanity check variable
    hbond_nbr = 0

    for col1col2 in bondlist:
        if col1col2[1] in hydroid2line.keys():
            #changing heavy atm mass
            line2modify = heavyid2line[col1col2[0]]
            former_mass = float(atomblock[line2modify][7])
            new_mass = former_mass + massdelta
            atomblock[line2modify][7] = str(new_mass)
            #changing hydrogen mass
            line2modify = hydroid2line[col1col2[1]]
            atomblock[line2modify][7] = str(new_hydrogen_mass)
            hbond_nbr += 1
        elif col1col2[0] in hydroid2line.keys():
            line2modify = heavyid2line[col1col2[1]]
            #atom mass is in 8th column
            former_mass = float(atomblock[line2modify][7])
            new_mass = former_mass + massdelta
            atomblock[line2modify][7] = str(new_mass)
            #changing hydrogen mass
            line2modify = hydroid2line[col1col2[0]]
            atomblock[line2modify][7] = str(new_hydrogen_mass)
            hbond_nbr += 1

    if hbond_nbr != len(hydroid2line):
        error = ('Error: the number of parsed hydrogens in hydrogenid variable is not the same as '
                 'the number of hydrogen bonds in hbond_nbr. Something is wrong !')
        raise SystemExit(error)

    massout = sum([float(column[7]) for column in atomblock.values()])

    return atomblock, massout


def write(toplines, atomblock, out):
    """This function reformates the modified atomblock, merges the toplines and the formatted
       modified atomblock and finally writes the ouptput topologies"""

    atomblock_f = {}
    for linenbr, atmline_list in atomblock.items():
        if len(atmline_list) == 8:
            atomblock_f[linenbr] = (f'{atmline_list[0]:>7}{atmline_list[1]:>12}'
                                    f'{atmline_list[2]:>8}{atmline_list[3]:>8}'
                                    f'{atmline_list[4]:>8}{atmline_list[5]:>8}'
                                    f'{f"{float(atmline_list[6]):.11g}":>12}'
                                    f'{f"{float(atmline_list[7]):.11g}":>12}\n')
        else:
            junk = ' '+' '.join([str(elem) for elem in atmline_list[8::]])
            atomblock_f[linenbr] = (f'{atmline_list[0]:>7}{atmline_list[1]:>12}'
                                    f'{atmline_list[2]:>8}{atmline_list[3]:>8}'
                                    f'{atmline_list[4]:>8}{atmline_list[5]:>8}'
                                    f'{f"{float(atmline_list[6]):.11g}":>12}'
                                    f'{f"{float(atmline_list[7]):.11g}":>12}'
                                    f'{junk}\n')

    #merging toplines with atomblock: entire new topology file
    toplines.update(atomblock_f)

    #To do: make sanity check: all toplines.keys should be different and increasing from 0 to
    #len(toplines)
    with open(out, 'w') as outtop:
        for line_nbr in range(0, len(toplines)):
            outtop.write(toplines[line_nbr])


def run_heavyh(topol = "./*.top", massfactor = 3, revert = 0, suff = "_heavyH"):
    """
    

    Parameters
    ----------
    topol : TYPE, optional, str, path, or regular expresion
        DESCRIPTION. The default is "./*.top".
        Give input topology (.top and/or itp), default: ./*.top, 
        if several .top in\ncurrent dir or given, error will be raised.
        However you can give as many .itp as you want (+max 1 .top)
    massfactor : TYPE, optional, float
        DESCRIPTION. The default is 3. int
        Give the factor by which each hydrogen mass should be multiplied or
        divided if you use revert.
    revert : TYPE, optional, float
        DESCRIPTION. The default is 0.
        Used if you want to decrease H mass instead, you should give the H
        mass in\nyour topology as argument for this flag
    suff : TYPE, optional, str
        DESCRIPTION. The default is "_heavyH".
        Suffix for output topology file name.

    Raises
    ------
    SystemExit
        DESCRIPTION.

    Returns
    -------
    None.

    """
    topol = glob(topol)

    top = [top for top in topol if ".top" in top]
    itp = [itp for itp in topol if ".itp" in itp]

    #Manage only one .top/inputs, this simplifies the script.
    if len(top) > 1:
        error = ('Several .top files in the current directory or given, please use -p to precise'
                 ' which .top should be used (you can however give as many .itp as you want or even'
                 ' just one .itp as input if you want). Tip: if you want to process several .top,'
                 ' just use a bash for loop over your .top using this scripti.')
        raise SystemExit(error)

    if revert != 0:
        massfactor = 1/massfactor

    topolin_topolout = file_checker(top, itp, suff)

    #variable needed to NOT modifiy the name of .itp without H in the .top
    no_hydrogen = []

    for items in topolin_topolout.items():
        print(f"\nProcessing {items[0]}...\n")

        try:
            toplines, atomblock, hydrogen_mass, hydroid2line,   \
            heavyid2line, bondlist, massin = top_parser(items[0], revert)

            atomblock, massout = atom_block_modifier(hydrogen_mass, massfactor, bondlist,
                                                     hydroid2line, heavyid2line, atomblock)

            write(toplines, atomblock, items[1])

            print(f'Difference between overall mass in input '
                  f'and overall mass in output is: {massin-massout:.4f} (should be 0)'
                  f'\n(massin is: {massin:.7g}, massout is: {massout:.7g})')

        except NoHydroError:
            print(f'***CAREFUL*** No hydrogens were found in: {items[0]}. If you think it makes '
                  f'sense for this molecular entity to not have hydrogens then you'
                  f'can ignore what follows.\n'
                  f'Else possible reasons could be:\n1. The hydrogen masses in the input are not '
                  f'between 0.99 and 1.1, this means that the script needs to be optimized.'
                  f'\n2. Because the top formatting is unusual'
                  f', you should review the code in this case too.\n')
            no_hydrogen.append(items[0])

    #This replaces the name of the modified .itp in the .top file if .itp contains H
    if top != []:
        #case where .top does not contain H
        if top[0] in no_hydrogen:
            new_name = top[0].replace('.top', suff+'.top')
            with open(top[0], 'r') as intop, open(new_name, 'w') as outtop:
                for line in intop:
                    for i in topolin_topolout:
                        if i not in no_hydrogen and i in line:
                            line = line.replace(i, topolin_topolout[i])
                    outtop.write(line)
        #case where .top does not contain atom block
        if top[0] not in topolin_topolout:
            new_name = top[0].replace('.top', suff+'.top')
            with open(top[0], 'r') as intop, open(new_name, 'w') as outtop:
                for line in intop:
                    for i in topolin_topolout:
                        if i not in no_hydrogen and i in line:
                            line = line.replace(i, topolin_topolout[i])
                    outtop.write(line)
        #case where .top does contain H
        else:
            with open(topolin_topolout[top[0]], 'r') as outtop, open('temp.top', 'w') as temp:
                for line in outtop:
                    for i in topolin_topolout:
                        if i not in no_hydrogen and i in line:
                            line = line.replace(i, topolin_topolout[i])
                    temp.write(line)
            #print('moving ./temp.top to', "./"+topolin_topolout[top[0]])
            move('./temp.top', "./"+topolin_topolout[top[0]])    



# If called from the command line...
if __name__ == "__main__":
    
    """This is just the main function of the script executing top_parser, atom_block_modifier
    and write functions, see __doc__ for details"""

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-p',
                        default=glob('./*.top'),
                        help='Give input topology (.top and/or itp), default: ./*.top, '
                             'if several .top in\ncurrent dir or given, error will be raised.'
                             ' However you can give as many .itp as you want (+max 1 .top)',
                        dest='topol',
                        nargs='*')
    parser.add_argument('-factor',
                        default=3,
                        help='Give the factor by which each hydrogen mass should be multiplied or'
                             ' divided if you use the -revert flag'
                             ' (default: %(default)s)',
                        dest='massfactor',
                        type=float)
    parser.add_argument('-revert',
                        default=0,
                        help='Used if you want to decrease H mass instead, you should give the H'
                             ' mass in\nyour topology as argument for this flag',
                        dest='revert',
                        type=float)
    parser.add_argument('-o',
                        default='_heavyH',
                        help=f'Suffix for output topology file name (default: %(default)s)',
                        action='store',
                        dest='suff',
                        type=str)
    args = parser.parse_args()

    top = [top for top in args.topol if ".top" in top]
    itp = [itp for itp in args.topol if ".itp" in itp]

    #Manage only one .top/inputs, this simplifies the script.
    if len(top) > 1:
        error = ('Several .top files in the current directory or given, please use -p to precise'
                 ' which .top should be used (you can however give as many .itp as you want or even'
                 ' just one .itp as input if you want). Tip: if you want to process several .top,'
                 ' just use a bash for loop over your .top using this scripti.')
        raise SystemExit(error)

    if args.revert != 0:
        args.massfactor = 1/args.massfactor

    topolin_topolout = file_checker(top, itp, args.suff)

    #variable needed to NOT modifiy the name of .itp without H in the .top
    no_hydrogen = []

    for items in topolin_topolout.items():
        print(f"\nProcessing {items[0]}...\n")

        try:
            toplines, atomblock, hydrogen_mass, hydroid2line,   \
            heavyid2line, bondlist, massin = top_parser(items[0], args.revert)

            atomblock, massout = atom_block_modifier(hydrogen_mass, args.massfactor, bondlist,
                                                     hydroid2line, heavyid2line, atomblock)

            write(toplines, atomblock, items[1])

            print(f'Difference between overall mass in input '
                  f'and overall mass in output is: {massin-massout:.4f} (should be 0)'
                  f'\n(massin is: {massin:.7g}, massout is: {massout:.7g})')

        except NoHydroError:
            print(f'***CAREFUL*** No hydrogens were found in: {items[0]}. If you think it makes '
                  f'sense for this molecular entity to not have hydrogens then you'
                  f'can ignore what follows.\n'
                  f'Else possible reasons could be:\n1. The hydrogen masses in the input are not '
                  f'between 0.99 and 1.1, this means that the script needs to be optimized.'
                  f'\n2. Because the top formatting is unusual'
                  f', you should review the code in this case too.\n')
            no_hydrogen.append(items[0])

    #This replaces the name of the modified .itp in the .top file if .itp contains H
    if top != []:
        #case where .top does not contain H
        if top[0] in no_hydrogen:
            new_name = top[0].replace('.top', args.suff+'.top')
            with open(top[0], 'r') as intop, open(new_name, 'w') as outtop:
                for line in intop:
                    for i in topolin_topolout:
                        if i not in no_hydrogen and i in line:
                            line = line.replace(i, topolin_topolout[i])
                    outtop.write(line)
        #case where .top does not contain atom block
        if top[0] not in topolin_topolout:
            new_name = top[0].replace('.top', args.suff+'.top')
            with open(top[0], 'r') as intop, open(new_name, 'w') as outtop:
                for line in intop:
                    for i in topolin_topolout:
                        if i not in no_hydrogen and i in line:
                            line = line.replace(i, topolin_topolout[i])
                    outtop.write(line)
        #case where .top does contain H
        else:
            with open(topolin_topolout[top[0]], 'r') as outtop, open('temp.top', 'w') as temp:
                for line in outtop:
                    for i in topolin_topolout:
                        if i not in no_hydrogen and i in line:
                            line = line.replace(i, topolin_topolout[i])
                    temp.write(line)
            #print('moving ./temp.top to', "./"+topolin_topolout[top[0]])
            move('./temp.top', "./"+topolin_topolout[top[0]])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
===============================================================================
Created on    : Fri Jun 25 21:00:10 2021
Author        : Alejandro Martínez León
Mail          : [alejandro.martinezleon@uni-saarland.de, ale94mleon@gmail.com]
Affiliation   : Jochen Hub's Biophysics Group
Affiliation   : Faculty of NS, University of Saarland, Saarbrücken, Germany
===============================================================================
DESCRIPTION   :
DEPENDENCIES  :
===============================================================================
"""
def warn(*args, **kwargs):
    pass
import warnings
warnings.warn = warn

import os, subprocess, shutil, datetime, tempfile, time, tqdm, pickle
from sklearn import cluster
from collections import OrderedDict
import numpy as np
import pandas as pd
import multiprocessing as mp
from glob import glob
from kmodes.kmodes import KModes

import prolif as plf
import MDAnalysis as mda
from MDAnalysis.topology.guessers import guess_types

from rdkit.Chem import DataStructs
from rdkit.DataStructs import ExplicitBitVect

from kneed import KneeLocator
from sklearn.cluster import KMeans
from sklearn.pipeline import Pipeline
from sklearn.metrics import silhouette_score
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from scipy.cluster.vq import vq
from scipy.interpolate import interp1d
from palmiche.utils import xvg




#=======================================================================================

#                          Miscellanea tools

#=======================================================================================
class CTE:
    """Some importnat physical constants
    """
    Kb = 8.314462618E-3 #kJ/(mol⋅K) (kNA)

def RSS(data, AlignPoint, ref_index_data = 0, NumbPoints = None, InterpolationKind = 'cubic'):
    """This function will return a list of Residues Squared Sums.

    Args:
        data (list): A list of np.arrays build in such a way that if reference = data[0]. Then for the reference: reference[:,0] give the independent variable and reference[:,1] the dependent one. NAd the same for the other entrances The first entrance of this list must be the reference
        AlignPoint (_type_): _description_
        NumbPoints (int, optional): _description_. Defaults to 1000.
        InterpolationKind (str, optional): _description_. Defaults to 'cubic'.

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    # Example:
    # In [1]: from palmiche.utils import tools, xvg
    # In [2]: d0 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord0_selected.xvg').data
    # In [3]: d1 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord1_selected.xvg').data
    # In [4]: d2 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord2_selected.xvg').data
    # In [5]: d3 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord3_selected.xvg').data
    # In [6]: d4 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord4_selected.xvg').data
    # In [7]: d5 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord5_selected.xvg').data
    # In [8]: data = [d0, d1, d2, d3, d4, d5]
    # In [9]: AlignPoint = 4.5
    # In [10]: NumbPoints = 300
    # In [11]: print(tools.RSS(data,AlignPoint, NumbPoints = NumbPoints))
    # [0.0, 962.7853017103384, 4159.864180622066, 9456.312066256418, 2076.1281115807665, 4421.657283337928]
    
    # GEtting the number of points for the interpolation
    if not NumbPoints:
        NumbPoints = data[0].shape[0]
    # Setting the reference
    ref = data[ref_index_data]
    # Getting the interpolation data
    f_interp_data = [interp1d(ref[:,0], item[:,1], kind=InterpolationKind) for item in data]
    
    minimum = max([item[:,0].min() for item in data])
    maximum = min([item[:,0].max() for item in data])
    # In this way we ensure only take those values for which we already have points in both data set

    if  not minimum <=AlignPoint <= maximum:
        raise ValueError(f"AlignPoint = {AlignPoint} is not in the range of the data, which is : {(minimum, maximum)}")
    
    # Interpolating data nad having the same points (newx)
    newx = np.linspace(minimum, maximum, num=NumbPoints, endpoint=True)
    interp_data = [f_interp(newx) for f_interp  in f_interp_data]


    #Aligning all the data
    for i in range(len(data)):
        if i == ref_index_data:
            continue
        else:
            offset =  f_interp_data[i](AlignPoint) - f_interp_data[ref_index_data](AlignPoint)
            interp_data[i] -= offset
    # import matplotlib.pyplot as plt
    # for i in range(len(data)):
    #     plt.scatter(data[i][:,0], data[i][:,1], label = f'row_data{i}')
    #     plt.plot(newx, interp_data[i],'->', label = f'interp_data{i}', ms = 4)
    # plt.legend()
    # plt.show()
    # Getting RSS values
    RSS_values = []
    for i in range(len(interp_data)):
        RSS_values.append(((interp_data[i] - interp_data[ref_index_data])**2).sum())
    
    return RSS_values

def RUS(data, ref_index_data = 0, NumbPoints = None, InterpolationKind = 'cubic'):
    """
    RUS stands for Residues Unsigned Sums
    RUS = sum_{i}^{n} (M_i - R_i)
    Where M_i is the error value of the magnitude on i of the model and R_i for the reference.
    RUS < 0, The model presents lower errors than the reference (in general)
    RUS = 0, The model has the same errors as the reference or they cancell out
    RUS > 0, The model presents worst errors than the reference
    """ 
    # GEtting the number of points for the interpolation
    if not NumbPoints:
        NumbPoints = data[0].shape[0]
    # Setting the reference
    ref = data[ref_index_data]
    # Getting the interpolation data
    f_interp_data = [interp1d(ref[:,0], item[:,1], kind=InterpolationKind) for item in data]
    
    minimum = max([item[:,0].min() for item in data])
    maximum = min([item[:,0].max() for item in data])
    # In this way we ensure only take those values for which we already have points in both data set

    
    # Interpolating data nad having the same points (newx)
    newx = np.linspace(minimum, maximum, num=NumbPoints, endpoint=True)
    interp_data = [f_interp(newx) for f_interp  in f_interp_data]

    # import matplotlib.pyplot as plt
    # for i in range(len(data)):
    #     plt.scatter(data[i][:,0], data[i][:,1], label = f'row_data{i}')
    #     plt.plot(newx, interp_data[i],'->', label = f'interp_data{i}', ms = 4)
    # plt.legend()
    # plt.show()
    # Getting RSS values
    RUS_values = []
    for i in range(len(interp_data)):
        RUS_values.append(((interp_data[i] - interp_data[ref_index_data])).sum())
    
    return RUS_values      



def zerolistmaker(n):
    return [0]*n
def unit_vector(vector):
    """ Returns the unit vector of the vector.  """
    return vector / np.linalg.norm(vector)

def angle_between(v1, v2):
    """ Returns the angle in radians between vectors 'v1' and 'v2'::

    # >>> angle_between((1, 0, 0), (0, 1, 0))
    # 1.5707963267948966
    # >>> angle_between((1, 0, 0), (1, 0, 0))
    # 0.0
    # >>> angle_between((1, 0, 0), (-1, 0, 0))
    # 3.141592653589793
    """
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))
def aovec(vectors, round = None):
    """Return the average oriented vector
    Calculate all the angles respect to the cartessina axis and average the angles,andthen return the unitary vector
    ang(vec, x) = mean(ang(vectors_i, x))
    ang(vec, y) = mean(ang(vectors_i, y))
    ang(vec, z) = mean(ang(vectors_i, z))
    vec = (cos(ang(vec, x); cos(ang(vec, y); cos(ang(vec, z))
    Args:
        vectors ([type]): [description]
        round ([type]): [description]
    Returns:
        [type]: [description]
    """
    vectors = [np.asarray(v) for v in vectors]
    #Dimension of the vectors, how many components
    vec = np.empty(vectors[0].shape)
    for i in range(vectors[0].shape[0]):
        zero_vector = np.zeros(vectors[0].shape)
        zero_vector[i] = 1
        vec[i] = np.cos(np.mean([angle_between(vector, zero_vector) for vector in vectors]))

    vec = unit_vector(vec)
    if round: vec = np.round(vec, round)
    return vec

def get_vec_COM(conf, tpr, ndx, group1, group2, round = None):
    """Take two groups of a index file and return the vector formed by the coordinates of group2, group1 in this order.py

    Args:
        conf (path): the configuration file [gro, pdb, xtc, etc..]
        tpr (path): tpr file, or gro, pdb, etc...
        ndx (path): index file, need to have the defined groups
        group1 (str): name of the first group in the index file.
        group2 (str): name of the second group in the index file.
        round (int or None): the specification to round.

    Returns:
        np.array: the unitary vector formed by group2, group1
    """
    xvgtmp1 = tempfile.NamedTemporaryFile(suffix=".xvg")
    run(f"export GMX_MAXBACKUP=-1;echo {group1} | gmx traj -f {conf} -s {tpr} -n {ndx} -com -ox {xvgtmp1.name}")
    coord1 = xvg.XVG(xvgtmp1.name).data[0,1:]# First row from column 1 to the end, the column 0 is time.

    xvgtmp2 = tempfile.NamedTemporaryFile(suffix=".xvg")
    run(f"export GMX_MAXBACKUP=-1;echo {group2} | gmx traj -f {conf} -s {tpr} -n {ndx} -com -ox {xvgtmp2.name}")
    coord2 = xvg.XVG(xvgtmp2.name).data[0,1:]# First row from column 1 to the end, the column 0 is time.
    vector = unit_vector(coord2 - coord1)
    if round: vector = np.round(vector, round)
    return vector


def timeit(method):
    def timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        if 'log_time' in kw:
            name = kw.get('log_name', method.__name__.upper())
            kw['log_time'][name] = int((te - ts) * 1000)
        else:
            print('%r  %2.2f ms' % \
                  (method.__name__, (te - ts) * 1000))
        return result
    return timed

#=======================================================================================

#                          Tools for execution

#=======================================================================================
def multi_run(commands, nPar, shell = True, executable = '/bin/bash'):
    """This will run as many runs as nPar.

    Args:
        commands (list): A list of string to be run in the specified shell.
        nPar (int): How many processes are running simultaneously
        shell (bool, optional): belongs to run(). Defaults to True.
        executable (str, optional): belongs to tun(). Defaults to '/bin/bash'.
        Popen (bool, optional): belongs to run(). Defaults to False.
    """

    tmp_cmds = []
    for cmd in commands:
        tmp_cmds.append(run(cmd, shell = shell, executable = executable, Popen = True))
        if len(tmp_cmds) >= nPar:
            print(f"{len(tmp_cmds)} running, now waiting...")
            for tmp_cmd in tmp_cmds:
                tmp_cmd.wait()
            tmp_cmds = []
        

def run(command, shell = True, executable = '/bin/bash', Popen = False):
    #Here I could make some modification in order that detect the operator system
    #NAd make the command compatible with the operator system
    #the function eval could be an option if some modification to the variable command 
    #need to be done.... SOme flight ideas...

    if Popen:
        #In this case you could access the pid as: run.pid
        process = subprocess.Popen(command, shell = shell, executable = executable, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text = True)
    else:
        process = subprocess.run(command, shell = shell, executable = executable, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text = True)
        returncode = process.returncode
        if returncode != 0:
            print(f'Command {command} returned non-zero exit status {returncode}')
            raise RuntimeError(process.stderr)
    return process

class CHECK:
    """
    This class test the normal ending of a GROMACS simulation using the log file.
    E.g. 
    let say that we have the md.log of a simulation
    # CHECK = CHECK('md.log')
    # if CHECK.performance:
    #     #The simulation had a good ending
    # else:
    #     #The simulations crashed.
    """

    def __init__(self, logfile, lisfile = None):
        self.logfile = logfile
        self.lisfile = lisfile
        self.performance = None  #in ns/day, this attribute could be used to check if the simulation ended
        self.__check_normal_end__()
    
    def __check_normal_end__(self):
        with open(self.logfile, 'r') as f:
            lines = f.readlines()
        for i in range(len(lines)):
            if 'Finished mdrun on rank' in lines[i]:
                try: 
                    self.performance = float(lines[i - 1].split()[1])
                except:
                    pass
                break

    def estimated_end_datetime(self):
        """Read an md.lis looking for the last instance of the string with the estimation of the ended time.
                "step 24940000, will finish Tue Oct  5 10:38:53 2021"

        Args:
            self.lisfile (path): md.lis, the output of the gromacs execution.

        Returns:
            datetime: The estimated end of the simulation.
        """
        with open(self.lisfile, "r") as f:
            lines = f.readlines()
        datetime_string = None
        for line in reversed(lines):
            if "will finish" in line:
                datetime_string = line.split("will finish")[-1].strip()
                break
                
        if datetime_string:
            return datetime.datetime.strptime(datetime_string, "%a %b %d %H:%M:%S %Y")
        else:
            return None
    
    def get_startdatetime_from_log(self):
        with open(self.logfile, "r") as f:
            lines = f.readlines()
        datetime_string = None
        for line in reversed(lines):
            if 'Started mdrun on rank' in line: # Will take the last
                datetime_string = " ".join(line.split()[-5:])
                break
                
        if datetime_string:
            return datetime.datetime.strptime(datetime_string, "%a %b %d %H:%M:%S %Y")
        else:
            return None               

    def daysdiff(self, startdatetime = None):
        """Take the difference between the estimated ended time of lisfile and datetime_start

        Args:
            lisfile (path): [description]
            datetime_start (datetime): A date time.

        Returns:
            float: Days' difference.
        """
        if not startdatetime:
            startdatetime = self.get_startdatetime_from_log()
            print("The starting data time of the log file was used becasue user did not provided a start_datetime")
        
        end_datetime = self.estimated_end_datetime()
        if startdatetime and end_datetime:
            return (end_datetime - startdatetime).total_seconds() / 86400
        else:
            return None

def checkrun(user = '$USER', partition = 'deflt'):
    """Check for the running process on the cluster.

    Returns:
        list of integers: The integers that identify the running process on the cluster.
    """
    stdout = run(f"squeue -p {partition} -u {user} --format=%.i").stdout.split()[1:]#The first line is the string "JOBID"

    return [int(item) for item in stdout]

def job_launch_list(job_path_list, shell="sbatch"):
    """
    Same as job_launch, but the path to the jobs are provided
    
    Args:
    job_path_list (_type_): _description_
    shell (str, optional): _description_. Defaults to "sbatch".

    Returns:
    _type_: _description_
    """
    cwd = os.getcwd()
    JOBIDs = []
    for job_path in job_path_list:
        path = os.path.dirname(job_path)
        script_name = os.path.basename(job_path)
        os.chdir(path)
        if shell == "sbatch":
            JOBIDs.append(int(run(f"{shell} --parsable {script_name}").stdout))
        else:
            try:
                JOBIDs.append(int(run(f"{shell} {script_name}").stdout))
            except:
                pass
    os.chdir(cwd)
    return JOBIDs

def launch_wait_check_repeat(partition, jobpaths, logfile, lisfile, jobsh_name, logfile_start_datetime = None):
    maximum_days = {
        'deflt': 2,
        'long': 5,
        'debug': 0.02,
        'medium': 2,
        'fat+': 2,
        'fat': 2,
        'int': 2,
        'gpu': 2,
        'gpu:test': 0.02,
        'gpu-hub':2,
            }
    # partition = 'deflt'
    # jobs2launch = [full_path_jobsh]
    # logfile = "production.log"
    # lisfile = 'production.lis'
    # jobsh_name = 'job.sh'
    # First launch (normal)
    JOBIDs = job_launch_list([os.path.join(path, jobsh_name) for path in jobpaths])
    while len(set(JOBIDs).difference(checkrun(partition=partition))) != len(JOBIDs):
        time.sleep(np.random.randint(20,60))
    # Relaunch if needed.
    jobs2relaunch = {}
    for path in jobpaths:
        check = CHECK(os.path.join(path,logfile), os.path.join(path,lisfile))
        # Check from the correct end
        print(f"The simulation {os.path.join(path,logfile)} presented a performance of {check.performance}")
        if not check.performance:
            if logfile_start_datetime:
                tmp_check = CHECK(os.path.join(path,logfile_start_datetime))
                daysdiff = check.daysdiff(tmp_check.get_startdatetime_from_log())
            else:
                daysdiff = check.daysdiff()
            # Check if the estimated time is larger than the reservation time in the corresponded partition.
            if daysdiff:
                if daysdiff > maximum_days[partition]:
                    # Get how many time the simulation must be relaunch
                    repeat = round((daysdiff / maximum_days[partition]) - 1)
                    if repeat == 0:repeat = 1
                    # Add path as key and number of repetition as value
                    jobs2relaunch[path] = repeat
                else:
                    print(f'We were not able to guess what is the problem with the run of the job {path}')
            else:
                print(f'We were not able to guess what is the problem with the run of the job {path}')
    # Only if there are some jobs to relaunch
    if jobs2relaunch:
        while sum(jobs2relaunch.values()):
            # Launch the jobs if repeat is greater than zero
            to_launch = [os.path.join(path, jobsh_name) for path in jobs2relaunch if jobs2relaunch[path] > 0]
            JOBIDs = job_launch_list(to_launch)
            while len(list(set(JOBIDs).difference(checkrun(partition=partition)))) != len(JOBIDs):
                time.sleep(np.random.randint(20,60))
            # Update the repetitions
            for path in jobs2relaunch:
                check = CHECK(os.path.join(path,logfile), os.path.join(path,lisfile))
                if check.performance:
                    jobs2relaunch[path] == 0
                if jobs2relaunch[path] > 0:
                    jobs2relaunch[path] -= 1
        print(f'Performance of the relaunched jobs:')
        for path in jobs2relaunch:
            check = CHECK(os.path.join(path,logfile), os.path.join(path,lisfile))
            print(f"The simulation {os.path.join(path,logfile)} presented a performance of {check.performance}")


def job_launch(shell="sbatch", script_name = "job.sh"):
    """
    

    Parameters
    ----------
    shell : TYPE, optional
        DESCRIPTION. The default is "sbatch".
    script_name : TYPE, optional
        DESCRIPTION. The default is "job.sh".
        If a regular expresion is provided
        then the function will execute the first ordered alphabetically. E.g:
        job.* was provided and there job.sh and job.bash. Then it will use job.bash.


    Returns
    -------
    JOBIDs : list of integers.
        DESCRIPTION. The ID of the launch in case of sbatch was used as shell

    """
    cwd = os.getcwd()
    JOBIDs = []
    for (root, _, _) in list(os.walk(cwd)):
        os.chdir(root)
        if glob(script_name):
            script_name = sorted(glob(script_name))[0]
            if shell == "sbatch":
                JOBIDs.append(int(run(f"{shell} --parsable {script_name}").stdout))
            else:
                try:
                    JOBIDs.append(int(run(f"{shell} {script_name}").stdout))
                except:
                    pass
    os.chdir(cwd)
    return JOBIDs


def backoff(file_path):
    """
    

    Parameters
    ----------
    file : TYPE string
        DESCRIPTION: The name or the path f     or the specific file

    Returns
    -------
    None.
    If the file already exist. it will made a back up to ./#{file}.{str(i)}#,
    Where i is an integer.
    """
    basname = os.path.basename(file_path)
    dirname = os.path.dirname(file_path)
    if os.path.exists(file_path):
        new_basname = basname
        i = 1
        while(os.path.exists(os.path.join(dirname, new_basname))):
            new_basname = f"./#{basname}.{str(i)}#"
            i += 1
        print(f"Back Off! I just backed up {file_path} to {os.path.join(dirname, new_basname)}")
        shutil.copy2(file_path, os.path.join(dirname, new_basname))


#=======================================================================================

#                          Tools for working with files

#=======================================================================================
def makedirs(path):
    if os.path.exists(path):
        pass
    else:
        os.makedirs(path,exist_ok=True)

def rm(pattern, r = False):
    """
    

    Parameters
    ----------
    patterns : string
        input-like Unix rm
    r : TYPE, bool
        DESCRIPTION. The default is False.
        If True delete also directories
    Returns
    -------
    None.

    """
    if glob(pattern):
        for p in glob(pattern):

            if r:
                if os.path.isdir(p):
                    shutil.rmtree(p)
                elif os.path.isfile(p):
                    os.remove(p)       
            else:
                if os.path.isfile(p):
                    os.remove(p)
                elif os.path.isdir(p):
                    print(f"The directory '{p}' was not deleted, set the flag r to True")
    else:
        
        print(f"rm: '{pattern}' doesn't exist")

                
def cp(src, dest, r = False):
    """
    This function makes use of the possible multiple CPU of the machine.

    Parameters
    ----------
    src : TYPE: string
        DESCRIPTION:
        Source Path to a directory or a file. regular expresion are accepted. The librery glob is used for that.
    dest : TYPE: string
        DESCRIPTION:
        Destination Path
    r : TYPE, optional
        DESCRIPTION. The default is False:
        If True, Also directories will be copy. If not, and a directory was
        given as src, a Raise Exception will be printed
        Another Raise Exception will be printed if the destination path doesn't
        exist.

    Returns
    -------
    None.

    """
    src = os.path.abspath(src)
    dest = os.path.abspath(dest)
    #This scheme doesn't consider the possibilities of other item except files, dir or regualr expresions
    if glob(src) and os.path.exists(os.path.dirname(dest)):
    
        if os.path.isdir(src):
            if r == True:            
                
                basename_src = os.path.basename(src)
                for root, dirs, files in os.walk(src):
                    path2copy = root.split(src)[1]
                    try:
                        if path2copy[0] == "/":
                            path2copy = path2copy[1:]
                    except: pass
                    path_dest = os.path.join(dest,basename_src,path2copy)
                    
                    if dirs:
                        pool = mp.Pool(mp.cpu_count())
                        pool.map(makedirs, [os.path.join(path_dest, d) for d in dirs])
                        pool.close()
                    if files:
                        makedirs(path_dest)
                        pool = mp.Pool(mp.cpu_count())
                        
                        pool.starmap(shutil.copy2, [(os.path.join(root, f),path_dest) for f in files])
                        pool.close()
            else:
                print(f"If you need to copy directories, set the 'r' flag to True. {src} is a directory")
    
        elif os.path.isfile(src):
            shutil.copy2(src,dest)
            #if not os.path.exists(dest):
             #   print(f"The file '{src} was not copy to '{dest}' becasue doesn't exist or is not accesible")
        
        
        else:#Here we are in regular expresions
            to_copy = [file for file in glob(src) if os.path.isfile(file)]
            pool = mp.Pool(mp.cpu_count())
            pool.starmap(shutil.copy2, [(pattern,dest) for pattern in to_copy])
            pool.close()
        
    elif not glob(src) and not os.path.exists(os.path.dirname(dest)):
        print(f"cp: neither {src}, nor {dest} exist.")
    elif not glob(src):
        print(f"cp: The source file {src} doesn't exist")
    else:
        print(f"cp: cannot create regular file '{dest}': Not a directory")
            
        

def mv(src, dest, r = False):
    cp(src, dest, r = r)
    rm(src,r=r)

def list_if_dir(path = '.'):
    return [item for item in os.listdir(path) if os.path.isdir(os.path.join(path, item))]

def list_if_file(path = '.'):
    return [item for item in os.listdir(path) if os.path.isfile(os.path.join(path, item))]

#=======================================================================================

#                      Tools for working with index and topology files

#=======================================================================================
def get_atom_index(file_path, H_atoms = True):
    """

    Parameters
    ----------
    file_path : TYPE, str
        DESCRIPTION. The file with the atoms, Could be an itp, gro or pdb.
    H_atoms : TYPE, optional
        DESCRIPTION. The default is True. :
            True:
                Return all the list of atoms.
            False:
                Just return the heavy atoms (non hydrogens).

    Returns
    -------
    atom_index : TYPE
        DESCRIPTION.

    """
    cwd = os.getcwd()
    tmpdir = tempfile.TemporaryDirectory()
    #Guessing the extension file
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"The file {file_path} doesn't exist or is not accessible")
    file_path = os.path.abspath(file_path)
    basename = os.path.basename(file_path)
    ext = basename.split('.')[-1].lower()
    os.chdir(tmpdir.name)
    atom_index = []
    if ext == 'itp':
        atoms_section = get_top_sections(file_path, dictionary=True)['atoms']
        for line in atoms_section:
            if not line.startswith(";") and not line == "\n":
                split = line.split()
                if H_atoms:
                    atom_index.append(int(split[0]))
                else:
                    if split[4].strip()[0] not in 'Hh': #This look for the first character of the atom name, if it is H or h, then is a hydrogen atom
                        atom_index.append(int(split[0]))
                 
    
    elif ext in ['gro', 'pdb']:
        with open("ndx.opt", "w") as opt:
            opt.write("""
                      "System" group System; 
                      "System-H" group System and ! name "H*" and ! name "h*"
                      """)
        run(f"""
                  echo 'q'| gmx make_ndx -f {file_path} -o tmp.ndx
                  gmx select -s {file_path} -sf ndx.opt -n tmp.ndx -on tmp
                  """)
        #deleting the line _f0_t0.000 in the file
        with open("tmp.ndx", "r") as index:
            data = index.read()
            data = data.replace("_f0_t0.000","")
        with open("tmp.ndx", "w") as index:
            index.write(data)
    
            ndx = NDX("tmp.ndx").data
            if H_atoms:
                atom_index = ndx["System"]
            else:
                atom_index = ndx["System-H"]
        

    # Returning to the original directory
    os.chdir(cwd)
    return atom_index
 # Both functions could be put together in a class NDX, that has one method, write, because always read. But this create the disatventage to incompatibiliti with ndx rectifivcation
 # En otras palabras que se pierde la ventaja de copiar un dictionario nen formatop NDX, se tendria que crear un NDX vacio (un archivo vacio , porque siempre lo va a leer) y asignar la data al objecto, 

class NDX:
    def __init__(self, file):
        self.file = file
        self.data = {}
        self.read()


    def read(self):
        """
        Da bateo cuando el nombre del indice no tiene ningu atomo, dime algo!!!!
        Tengo que revisar en todos lo lugares que llamo a NDX, porque le voy a quitar que convierta a entero
        Y tengo que ver como hago mas general la idea de indices sin atomos


        Returns
        -------
        dictionary : TYPE: dict
            A dictionary with the group index in the form
            { group_name1: [atom_idndex1, atom_idndex2, ...]
            group_name2: [atom_idndex1, atom_idndex2, ...]
            ...
                }

        """
        with open(self.file,'r') as f:
            lines = f.readlines()
               
        positions = [i for (i, line) in enumerate(lines) if line.startswith('[')]

        for i in range(len(positions)):
            key = lines[positions[i]][2:-3]
            self.data[key] = []
            try:
                chunk_lines = lines[positions[i]+1:positions[i+1]]
            except:
                chunk_lines = lines[positions[i]+1:]
            for line in chunk_lines:
                self.data[key] += [int(value) for value in line.split()]


    def write(self, out_name = "index.ndx", backup = True):
        """
        Write a GROMACS index file

        Parameters
        ----------
        dictionary : dict
            The keys of the dictionary are the names of the groups.
            The values are the corresponded list of atoms
        out_name : TYPE, str
            DESCRIPTION. The default is "index.ndx".
            The name of the output file

        Returns
        -------
        None.

        """
        data = ""
        items_per_row = 15
        for key in self.data:
            data += f"[ {key} ]\n"
            i = 0

            while i < len(self.data[key]):
                string = [str(item) for item in self.data[key][i:i+items_per_row]]
                i += items_per_row
                data += " ".join(string) + "\n"

        if backup: backoff(out_name)
        out_file = open(out_name,'w')
        out_file.write(data)
        out_file.close()


def get_top_sections(topology, dictionary = False):
    """
    The flag dictionary is available because if in the topology there are two
    sections with the same name, then the last one is the only one that will 
    be save.
    IN the case that 

    Parameters
    ----------
    topology : path to a gromacs topology file
        DESCRIPTION.
    dictionary : TYPE, optional
        DESCRIPTION. The default is False.
        If True instead of a list return a dict with keywords the sections
        If not, a list with elements [key, info], ...

    Returns
    -------
    sections : list or dict depend on dictionary
        DESCRIPTION.
        The info of the topology file

    """
    with open(topology,"r") as top:
        top_lines = top.readlines()
    
    sections = [] #is not possible to use dict because there is some repeated sections, like dihedrals, I don't know why
    for i in range(len(top_lines)):
        if top_lines[i].startswith("["):
            
            key = ""
            for char in top_lines[i]:
                if char == "]":
                    break
                if char not in "[ ":
                    key += char
            key.strip()
            
            
            info = []
            for j in range(i + 1, len(top_lines)):
                if top_lines[j].startswith("["):
                    i = j - 1
                    break
                info.append(top_lines[j])
               
            sections.append([key.strip(), info])   
    
    if dictionary: return dict(sections)
    return sections
def KbT(absolute_temperature):
    """Return the value of Kb*T in kJ/mol

    Args:
        absolute_temperature (float): The absolute temperature in kelvin
    """
    
    return absolute_temperature*CTE.Kb

def beta(absolute_temperature):
    return 1 / KbT(absolute_temperature)

def BoltzmannFactor(E,T):
    return np.exp(-E/KbT(T))

def get_force_cte(T, sigma):
    return KbT(T)/sigma**2 # Kj/(mol [sigma^2])


def get_sigma(T, force_constant):
    return np.sqrt(KbT(T)/force_constant)

#======================================================================================================
# Here I will introduce some nice tools for working with advances python modules for molecular dynamic
#======================================================================================================
def my_guess_types(names):
    """Guess the type of the atoms. It will simply remove the number of each string. Useful for the guessing of the ligands atoms

    Args:
        names (list): a list of atoms names
    """
    def f(array):
        for item in array:
            return ''.join(filter(str.isalpha, item))
    return np.vectorize(f)(names)

def array_to_bv(s):
    bv = ExplicitBitVect(len(s))
    on_bits = np.where(s == True)[0].tolist()
    bv.SetBitsFromList(on_bits)
    return bv

class classifier_ifp:
    def __init__(self, tpr, xtc, ligand_selection = 'resname LIA', protein_selection = 'protein', elapsed_steps = 0, load_fingerprint = False, ifp_file = 'ifp.pkl', ifp_atoms_file = 'ifp_atoms.pkl', n_clusters = 2) -> None:
        self.tpr = tpr
        self.xtc = xtc
        self.ligand_selection = ligand_selection
        self.protein_selection = protein_selection
        self.elapsed_steps = elapsed_steps
        self.load_fingerprint = load_fingerprint
        self.ifp_file = ifp_file
        self.ifp_atoms_file = ifp_atoms_file
        self.n_clusters = n_clusters
        self.parse()
    
    def __split(self, string):
        def f(s):
            if 'chain' in s:
                return s.split('_')[-1]
            elif "LI" in s:
                return s.split('_')[-1][-1]
            else:
                return ""
        return np.vectorize(f)(string)
    
    def cluster(self, n_clusters=2, max_iter=500, init='Cao', n_init=40, verbose=0, random_state=None, n_jobs=1):
        self.n_clusters = n_clusters
        ## Becasue all the data is True or False, every variable contribute with the same weight, so there is no need to rescale it.
        ## Initialize the model
        self.model = KModes(
            n_clusters=self.n_clusters,
            max_iter=max_iter,
            init=init,
            n_init=n_init,
            verbose=verbose,
            random_state=random_state,
            n_jobs=n_jobs
        )
        ## Run the model
        self.model.fit(self.ifp_df)
        # Group frames on clusters
        self.membership = {}
        for i in range(self.n_clusters):
            self.membership[i] = np.where(self.model.labels_ == i)[0]

    def parse(self):
        if self.load_fingerprint:
            self.ifp_df = pd.read_pickle(self.ifp_file).astype(np.uint8)
            try:
                self.ifp_df_atoms = pd.read_pickle(self.ifp_atoms_file)
            except:
                print('This object does not contains the atribute self.ifp_df_atoms')
        else:
            tmp_pdb = tempfile.NamedTemporaryFile(suffix='.pdb')
            run(f"export GMX_MAXBACKUP=-1;  gmx editconf -f {self.tpr} -o {tmp_pdb.name}")


            ref = mda.Universe(tmp_pdb.name) # This has the correct numeration for the residues
            u = mda.Universe(self.tpr, self.xtc)#, guess_bonds=True

            guessed_elements = guess_types(my_guess_types(u.atoms.names))
            u.add_TopologyAttr('names', guessed_elements)
            u.add_TopologyAttr('elements', guessed_elements)
            u.add_TopologyAttr('types', guessed_elements)
            u.add_TopologyAttr('resid', ref.residues.resids)
            u.add_TopologyAttr('chainID', self.__split(u.atoms.segids)) # Puede pensar en algo mas general con la funcion que tengo dentro del pdb para adivinar las cadenas

            self.ligand = u.atoms.select_atoms(self.ligand_selection)
            self.protein = u.atoms.select_atoms(self.protein_selection)

            # Initialize fingerprint
            ifp = plf.Fingerprint()
            # run on a slice of frames from begining to end with a step of 10
            if self.elapsed_steps:
                ifp.run(u.trajectory[::self.elapsed_steps], self.ligand, self.protein)
            else:
                ifp.run(u.trajectory, self.ligand, self.protein)

            self.ifp_df = ifp.to_dataframe()
            self.ifp_df_atoms = ifp.to_dataframe(return_atoms = True)

            # Saving the fingerprint
            self.ifp_df.to_pickle(self.ifp_file)
            # Also the ifp_atoms will be exported
            self.ifp_df_atoms.to_pickle(self.ifp_atoms_file)

        # Run cluster with the default values nad number of cluster
        self.cluster(n_clusters=self.n_clusters)

    def get_closers(self, user_membership = None):
        """Return the closest frames for each cluster in self.membership or user_memebership if provided. 

        Args:
            user_membership (dict, optional): The memebership, the keys will be the cluster identifier and the values the frames taht belongs to each cluster. Defaults to None.

        Returns:
            _type_: _description_
        """
        bvs_centroids = [array_to_bv(centroids) for centroids in self.model.cluster_centroids_]
        bvs_frames = plf.to_bitvectors(self.ifp_df)

        if user_membership:
            membership = user_membership
        else:
            membership = self.membership

        closest_frames = {}
        for cluster in membership:
            bvs_selected_frames = [bvs_frames[frame] for frame in membership[cluster]]
            # If a cluster only have one member
            if len(bvs_selected_frames) == 1:
                # BulkTanimotoSimilarity needs more than one fingerprint, this is just a trick, and probably alwasy be avoid.
                bvs_selected_frames += bvs_selected_frames
            similarities = DataStructs.BulkTanimotoSimilarity(bvs_centroids[cluster], bvs_selected_frames)
            index = similarities.index(max(similarities))
            closest_frames[cluster] = membership[cluster][index]
        return closest_frames

class classifier_dist_contacts:
    def __init__(self, tpr, xtc, chains = ('A', 'B', 'C', 'D', 'E'), elapsed_steps = 0, n_clusters = 2) -> None:
        self.tpr = tpr
        self.xtc = xtc
        self.chains = chains
        self.elapsed_steps = elapsed_steps
        self.n_clusters = n_clusters
        self.data = pd.DataFrame()
        self.parse()
    
    def cluster(self, n_clusters=2, init='k-means++', n_init=50, max_iter=500, tol=0.0001, verbose=0, random_state=None, copy_x=True, algorithm='auto'):
        features = ['AveNumResi', 'AveDistZ']
        self.n_clusters = n_clusters
        pipe = Pipeline(
            [
                ("scaler", MinMaxScaler()),
                ('kmeans',
                    KMeans(
                        n_clusters=self.n_clusters,
                        init=init,
                        n_init=n_init,
                        max_iter=max_iter,
                        tol=tol,
                        verbose=verbose,
                        random_state=random_state,
                        copy_x=copy_x,
                        algorithm=algorithm
                    )
                )
            ]
        )
        pipe.fit(self.data[features])
        self.model = pipe['kmeans']
        #Think about if implement all the method in a athomatic way to improve the selction of the clusters
        # Group frames on clusters
        self.membership = OrderedDict()
        for i in range(self.n_clusters):
            self.membership[i] = np.where(self.model.labels_ == i)[0]
    
    def parse(self):
        u = mda.Universe(self.tpr, self.xtc)

        close_contacts_frame_0 = {chain:u.atoms.select_atoms(f'(backbone and segid *Protein_chain_{chain}) and around 4 resname LI{chain}') for chain in self.chains}

        if self.elapsed_steps:
            iterator = u.trajectory[::self.elapsed_steps]
            Frame = range(0,u.trajectory.n_frames,self.elapsed_steps)
        else:
            iterator = u.trajectory
            Frame = range(u.trajectory.n_frames)
        AveNumResi = []
        AveDistZ = []
        for ts in tqdm.tqdm(iterator):
            NumResi = []
            DistZ = []
            for chain in self.chains:
                NumResi.append(len(u.atoms.select_atoms(f'(protein and segid *Protein_chain_{chain}) and around 3 resname LI{chain}').residues))
                DistZ.append(np.abs(close_contacts_frame_0[chain].center_of_mass(pbc=True, compound='group')[-1] - u.atoms.select_atoms(f'resname LI{chain}').center_of_mass(pbc=True, compound='group')[-1]))
            AveNumResi.append(np.mean(NumResi))
            AveDistZ.append(np.mean(DistZ))
        self.data = pd.DataFrame(
            {
                'Frame': Frame,
                'AveNumResi': AveNumResi,
                'AveDistZ': AveDistZ,
            }
        )
        # Run cluster with the default values nad number of cluster
        self.cluster(n_clusters=self.n_clusters)
    
    def get_closers(self):
         # https://stackoverflow.com/questions/26795535/output-50-samples-closest-to-each-cluster-center-using-scikit-learn-k-means-libr
        # This is like on the paper was done, but I need to do the scalling first
        scale = MinMaxScaler()
        scale.fit(self.data[['AveNumResi', 'AveDistZ']])
        closest_frames, distances = vq(self.model.cluster_centers_, scale.transform(self.data[['AveNumResi', 'AveDistZ']]))
        return closest_frames

if __name__ == '__main__':
    
    pass
    #k = aovec([(0,0,1), (0,1,0), (1,0,0)])
    #deg = 2.5
    #rad = np.pi*deg/180
    #n = 0.05
    #print(get_force_cte(298.15, n))
    #print(get_sigma(303.15, 500))
    # now = datetime.datetime.now()
    #ndx = NDX("/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_J/7e27/sys_MMV007839_Cell_891_SP_param/index.ndx")
    #print(ndx.data)
    # check = CHECK("md.log", "md.lis")
    # print(check.performance)
    # print(check.estimated_end_datetime())
    # print(check.daysdiff(now))
    #print(os.listdir())
    #cp("test.py", "h/k.py")
    #print(list_if_file("./"))q
    #print(os.path.isfile("./*py"))
    # tpr = '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_M_annealing/7e27/sys_MMV007839_Cell_891_SP_param/pull.tpr'
    # xtc = '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_M_annealing/7e27/sys_MMV007839_Cell_891_SP_param/pull.xtc'
    # fingerprint_files = [
    #     '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_M_annealing/7e27/sys_MMV007839_Cell_891_SP_param/ifp.pkl',
    #     '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_M_annealing/7e27/sys_MMV007839_Cell_891_SP_param/ifp_atoms.pkl'
    # ]
    # ligand = 'LIA'

    # a = classifier_dist_contacts(tpr, xtc, elapsed_steps=200)
    # #print(a.data)
    # #print(a)
    # #a = classifier_ifp(tpr, xtc, ligand_selection='resname LIA', elapsed_steps = 0, load_fingerprint = True, ifp_file = fingerprint_files[0], n_clusters = 10)
    # print(a.model.cluster_centers_.shape)
    # a.cluster(n_clusters=3)

    # print(a.model.cluster_centers_)
    # print(a.get_closers())
    # #print(a.get_closers())
    # #print(a.labels_)
    #print(KbT(300))
    d0 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord0_selected.xvg').data
    d1 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord1_selected.xvg').data
    d2 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord2_selected.xvg').data
    d3 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord3_selected.xvg').data
    d4 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord4_selected.xvg').data
    d5 = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_N_ST/7e27/sys_MMV007839_Cell_891_SP_param/windows/coord5_selected.xvg').data
    data = [d0, d1, d2, d3, d4, d5]
    AlignPoint = 4.5
    NumbPoints = 300
    print(RSS(data,AlignPoint, NumbPoints = NumbPoints))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import glob as glob
import multiprocessing as mp
import os
import numpy as np
from palmiche.utils import tools
import tqdm

#========Deffinitions of class and methosd to get the output of Vina===========
class Atom:
    #https://userguide.mdanalysis.org/stable/formats/reference/pdbqt.html#writing-out
    def __init__(self, line):
        self.lineType = "ATOM"
        self.serial = int(line[6:11])
        self.name = line[12:16].strip()
        self.altLoc = line[16]
        self.resName = line[17:21].strip()
        self.chainID = line[21]
        self.resSeq = int(line[22:26])
        self.iCode = line[26]
        self.x = float(line[30:38])
        self.y = float(line[38:46])
        self.z = float(line[46:54])
        self.occupancy = line[54:60].strip()
        self.tempFactor = line[60:66].strip()
        self.partialChrg = line[66:76].strip()
        self.atomType = line[78:80].strip()
    def __getitem__(self, key):
        return self.__dict__[key]

class Hetatm:
    def __init__(self, line):
        self.lineType = "HETATM"
        self.serial = int(line[6:11])
        self.name = line[12:16].strip()
        self.altLoc = line[16]
        self.resName = line[17:21].strip()
        self.chainID = line[21]
        self.resSeq = int(line[22:26])
        self.iCode = line[26]
        self.x = float(line[30:38])
        self.y = float(line[38:46])
        self.z = float(line[46:54])
        self.occupancy = line[54:60].strip()
        self.tempFactor = line[60:66].strip()
        self.partialChrg = line[66:76].strip()
        self.atomType = line[78:80].strip()
    def __getitem__(self, key):
        return self.__dict__[key]

class Remark:
    def __init__(self, line):
        pass

class CHUNK_VINA_OUT:
    def __init__(self, chunk):
        self.chunk = chunk
        self.atoms = []
        self.run = None
        self.freeEnergy = None
        self.RMSD1 = None
        self.RMSD2 = None
        self.parse()

    def parse(self):
        for line in self.chunk:
            if line.startswith("MODEL"):
                self.run = int(line[5:])
            elif line.startswith("REMARK VINA RESULT:"):
                    (self.freeEnergy, self.RMSD1, self.RMSD2) = [float(number) for number in line.split(":")[-1].split()]
                    
            elif line.startswith("ATOM"):
                self.atoms.append(Atom(line))
            else:
                pass

    def get_atoms(self):
        """Return a list of all atoms.

        If to_dict is True, each atom is represented as a dictionary.
        Otherwise, a list of Atom objects is returned."""
        return [x.__dict__ for x in self.atoms]

    def get_vector(self, atom_index1, atom_index2, normalized = True):
        """
        Parameters
        ----------
        atom_index1 : TYPE: int
            DESCRIPTION: The index of the first atom. Remember that python list
            start at 0. So, the atom number 65 in a pdbqt will be at the position
            64 of the list of atoms.
        atom_index2 : TYPE: int
            DESCRIPTION: The index of the second atom. Remember that python list
            start at 0. So, the atom number 1 in a pdbqt will be at the position
            0 of the list of atoms.
        normalized : TYPE, optional
            DESCRIPTION. The default is True.

        Returns
        -------
        a numpy array. the vector between the two atoms, the vector point to 
        to atom2 (atom1|----->atom2)
        """
        XYZ_atom1 = (self.atoms[atom_index1].x, self.atoms[atom_index1].y, self.atoms[atom_index1].z)
        XYZ_atom2 = (self.atoms[atom_index2].x, self.atoms[atom_index2].y, self.atoms[atom_index2].z)
        vector = np.array(XYZ_atom2) - np.array(XYZ_atom1)
        
        if normalized:
            vector /= np.linalg.norm(vector)
        return vector
        
    def write(self, name = None):
        if name:
            with open(name,"w") as f:
                f.writelines(self.chunk)
        else:
            with open(f"Run_{self.run}.pdbqt","w") as f:
                f.writelines(self.chunk)            

class VINA_OUT:
    """
    To acces the chunks you need to take into account that 
    """
    def __init__(self, file):
        self.file = file

        self.chunks = []
        self.parse()

    def parse(self):
        with open(self.file, "r") as input_file:
            lines = input_file.readlines()
        i = 0
        while i < len(lines):

            if lines[i].startswith("MODEL"):
                j = i
                tmp_chunk = []
                while (not lines[j].startswith("ENDMDL")) and (j < len(lines)):
                    tmp_chunk.append(lines[j])
                    j += 1
                    i += 1
                tmp_chunk.append("ENDMDL\n")
                self.chunks.append(CHUNK_VINA_OUT(tmp_chunk))

            i += 1
    def PosNegConf(self, atom_1, atom_2, vector = (0,0,1)):
        """
        

        Parameters
        ----------
        atom_1 : TYPE: int
            DESCRIPTION: The index of the atom 1 (you could know this using pymol, label>atomidentifier>ID )
        atom_2 : TYPE: int
            DESCRIPTION: The index of the atom 2 (you could know this using pymol, label>atomidentifier>ID )
        vector : TYPE, optional, is the reference vector in order to 
        look for the confomers in the direction and in the opposite direction 
        of this vector
            DESCRIPTION. The default is [0,0,1].

        Returns
        -------
        PO_chunk and NO_chunk obkjects and positive_oriented.pdbqt and negative_oriented.pdbq

        """
        
        positive_orientation = []
        negative_orientation = []
     
        
        for chunk in self.chunks:
            vector_chunk = chunk.get_vector(atom_1 - 1, atom_2 - 1, normalized=True)
            dot_product = np.dot(vector_chunk, vector)
            if dot_product >= 0:
                positive_orientation.append(chunk)
            else:
                negative_orientation.append(chunk)
        
        if positive_orientation:
            PO_chunk = min(positive_orientation,key=lambda x: x.freeEnergy)
            PO_chunk.write("positive_oriented.pdbqt")
        else:
            print("A positive orientaion was not found")
        if negative_orientation:    
            NO_chunk = min(negative_orientation,key=lambda x: x.freeEnergy)
            NO_chunk.write("negative_oriented.pdbqt")
        else:
            print("A negative orientaion was not found")
        
        return PO_chunk, NO_chunk
            
    def BestEnergy(self, write = True):
        min_chunk = min(self.chunks, key= lambda x: x.freeEnergy)
        if write: min_chunk.write("best_energy.pdbqt")
        
        return min_chunk
            
        
#==============================================================================





#====================Congfiguration============================================
def vinaconfig(receptor_path, ligand_path, grid_center, size, modes = 1, config_path = "./", chain = "A"):
    """
    

    Parameters
    ----------
     receptor : TYPE
        DESCRIPTION:
            path to the receptor (relative or absolute)
    
    ligand : TYPE
        DESCRIPTION:
            path to the ligand (relative or absolute)
            e.g:
                mol.pdbqts

    grid_center : TYPE
        DESCRIPTION:
            A tupla with the float coords of the center:
                (x,y,z)
    size : TYPE
        DESCRIPTION:
            A tupla with the size (float > 0) on each coord:
                (Sx,Sy,Sz)
            It is important to consider that the grid space used is grid_space = 0.375
            This is the default value used 
    modes : TYPE, optional
        DESCRIPTION:
            How many modes will be computed
            The default is 1. 

    Returns
    -------
    None.

    """


    name = os.path.basename(os.path.abspath(ligand_path)).split(".")[0]
    config_path = os.path.abspath(config_path)
    receptor_path = os.path.abspath(receptor_path)
    ligand_path = os.path.abspath(ligand_path)
    
    final=open(f"{config_path}/{name}_{chain}.config",'w')

    final.write(f"receptor      =  {receptor_path}\n")
    final.write(f"ligand        =  {ligand_path}\n")
    
    final.write(f"center_x      =  {grid_center[0]}\n")
    final.write(f"center_y      =  {grid_center[1]}\n")
    final.write(f"center_z      =  {grid_center[2]}\n")
    
    final.write(f"size_x        =  {size[0]}\n")
    final.write(f"size_y        =  {size[1]}\n")
    final.write(f"size_z        =  {size[2]}\n")
    final.write(f"num_modes     =  {modes}\n")
    
    final.write(f"out       =  {config_path}/{name}_{chain}_out.pdbqt\n")

    final.close()


#==============================================================================

#=================Deffinition of the run=======================================
def runvina(grid_opt, exhaustiveness = 25, cpu = 4, chains = ["A"], modes = 1, continue_check = False):
    """
    Parameters
    ----------
    grid_opt : TYPE
        DESCRIPTION.
    exhaustiveness : TYPE, optional
        DESCRIPTION. The default is 25.
    cpu : TYPE, optional
        DESCRIPTION. The default is 4.
    chains : TYPE, optional
        DESCRIPTION. The default is ["A"].

    Raises
    ------
    ValueError
        DESCRIPTION.

    Returns
    -------
    None.

    """
    cwd = os.getcwd()
    
    
    ligand_paths = glob.glob(os.path.join(cwd, "ligands", "*.pdbqt"))
    ligands_name = [os.path.basename(ligand_path).split(".")[0] for ligand_path in ligand_paths]
    
    receptor_paths = glob.glob(os.path.join(cwd, "receptors", "*.pdbqt"))
    receptors_name = [os.path.basename(receptor_path).split(".")[0] for receptor_path in receptor_paths]
    
    cmd = []
    for i, receptor_name in enumerate(receptors_name):
        
        tools.makedirs(f"{cwd}/run_vina_run/{receptor_name}")
        for j, ligand_name in enumerate(ligands_name):
            tools.makedirs(f"{cwd}/run_vina_run/{receptor_name}/{ligand_name}")
            for chain in chains:
                tools.makedirs(f"{cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}")
                #print(f"{cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}")
                #tools.cp(receptor_paths[i], f"{cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}")
                #tools.cp(ligand_paths[j], f"{cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}")
                try:
                    param = grid_opt[receptor_name][chain]
                    
                    vinaconfig(receptor_paths[i],ligand_paths[j],
                                               param["boxcenter"],
                                               param["boxsize"],
                                               config_path=f"{cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}",
                                               chain=chain,
                                               modes=modes)
                    #Check for the _out.pdbqt, this is only exported when the calculation finished
                    if continue_check:
                        if not os.path.isfile(f"{cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}/{ligand_name}_{chain}_out.pdbqt"):
                            cmd.append(f"vina --config {cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}/{ligand_name}_{chain}.config --exhaustiveness={exhaustiveness} --cpu={cpu}")
                    else:
                        cmd.append(f"vina --config {cwd}/run_vina_run/{receptor_name}/{ligand_name}/{chain}/{ligand_name}_{chain}.config --exhaustiveness={exhaustiveness} --cpu={cpu}")
                except:
                    print(f"{receptor_name} or/and {chain} of {receptor_name} is/are not present(s) in {grid_opt}")
    
    
    #Parallel running
    #Creating the iterable was already created (cmd)
    #[print(c) for c in cmd]
    #Creating the pool
    pool = mp.Pool(int(mp.cpu_count()/cpu))
    #Running
    for i in tqdm.tqdm(pool.imap_unordered(tools.run, cmd), total=len(cmd)):
        pass
    
    #Closing
    pool.close()
#==============================================================================



if __name__ == '__main__':
    pass
    #vina = VINA_OUT("mol1_out.pdbqt")
    #vina.PosNegConf(atom_1 = 17 , atom_2 = 1)
    #runvina("/home/ale/MY_PYTHON_PACKEGES/palmiche/examples/Vina_Docking/ref_info/boxes.box")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from palmiche.utils import tools, xvg
import argparse
import multiprocessing as mp
import os, re, tqdm
from scipy import integrate
import numpy as np


def gmx_analyze(f:str, ac:str = 'autocorr.xvg', overwrite:bool = False, reaction_coordinates:list = None, **kwargs):
    """Execute gmx analyze command

    Parameters
    ----------
    f : str
        The path to the xvg file with the format (time, value)
    ac : str, optional
        The name of the output for the autocorrelation calculations, by default 'autocorr.xvg'
    overwrite : bool, optional
        If True and if the ac file exists, this file will be overwrite,
        if False it will not perform any new calculation, by default False
    reaction_coordinates : list, optional
        This is a list for the reaction coordinate to take into account.
        If f has six columns, the first one will be considered time and the rest the reaction coordinates.
        If we would like to calculate for the second and third columns, then we must provided the list [2,3]
        by default None
    kwargs : Extra parameters will be passed directly to gmx analyze.
    """
    print(ac)
    if reaction_coordinates:
        columns = [0] + reaction_coordinates # we must pug the time
        f_dirname = os.path.dirname(f)
        f_name = os.path.basename(f).split('.')[0]
        f2work = os.path.join(f_dirname, f"{f_name}_acf_processed.xvg")
        f_xvg = xvg.XVG(f)
        f_xvg.data = f_xvg.data[:,columns]
        f_xvg.write(f2work)
    else:
        f2work = f

    if overwrite:
        cmd = 'export GMX_MAXBACKUP=-1; '
    else:
        cmd = ''
    cmd += f'gmx analyze -f {f2work} -ac {ac} '
    for key in kwargs:
        cmd += f'-{key} {kwargs[key]} '
    if not os.path.isfile(ac) or overwrite:
        tools.run(cmd)

def ___gmx_analyze_wrapper___(kwargs):
    """A simple wrapper to use parallelization of gmx_analyze.
    """
    gmx_analyze(**kwargs)

def IntegralOfACF(windows_path:str,  ac:str = 'autocorr.xvg', integrator:str = 'trapezoid', njobs:int = 1):
    """This will calculate the integral of the autocorrelation function

    Parameters
    ----------
    windows_path : str
        The path were the windows directories are with the format of of five integers. E.g. for window 1: 00001
    ac : str, optional
        The name of the output for the autocorrelation calculations, by default 'autocorr.xvg'
    integrator : str, optional
        Integration method. Are currently supported:
            trapezoid --> scipy.integrate.trapezoid
            trapz --> numpy.trapz (Default)
            simps --> scipy.integrate.simpson
    njobs : int, optional
        The number of jobs for the parallelization, by default 1

    Returns
    -------
    np.array of shape (number of windows X number of reaction coordinates)
    and a txt file with the saved dat in windows_path directory.

    Raises
    ------
    ValueError
        If an invalid integrator is provided.
    """
    pattern = re.compile("^\d{5}$")
    individual_paths = sorted([os.path.join(windows_path, window) for window in tools.list_if_dir(windows_path) if pattern.match(window)])

    # Selecting the integrator method
    if integrator == 'trapezoid':
        integrator = integrate.trapz
    elif integrator == 'trapz':
        integrator = np.trapz
    elif integrator == 'simpson':
        integrator = integrate.simps
    else:
        raise ValueError(f"The integration method '{integrator}' is not available.\n"\
            "Current possible selections are:\n"\
            "   -> trapezoid\n"\
            "   -> trapz\n"\
            "   -> simpson (Default)")

    # Getting all the data
    all_data = []
    for window in individual_paths:
        multidata = xvg.XVG(os.path.join(window, ac)).multidata

        data_on_window = multidata[0]
        for data in multidata[1:]:
            data_on_window = np.insert(data_on_window,-1, data[:,1], axis = -1)
        all_data.append(data_on_window)
    #all data shape = (number of windows, number of time points, number or reaction coordinates +1 [time is the first column])
    all_data = np.stack(all_data, axis = 0)

    # Get the IACFs per windows
    args_per_window = []
    for i in range(all_data.shape[0]):
        args_per_window.append((all_data[i], integrator))

    pool = mp.Pool(min(njobs, mp.cpu_count()))
    integrals_per_window = pool.starmap(___IntegralOfACF_chunk___, args_per_window)
    pool.close()
    integrals_per_window = np.asarray(integrals_per_window)
    # Saving data
    np.save(os.path.join(windows_path, 'IACFs_per_windows.npy'), integrals_per_window)

    # Get the integral on the time projection
    # Rearrange the data
    rearranged_data = []
    for j in range(all_data.shape[1]):
        rearranged_data.append(all_data[:,j,:])
    rearranged_data = np.asarray(rearranged_data)

    args_per_time = []
    for i in range(rearranged_data.shape[0]):
        args_per_time.append((rearranged_data[i], integrator))

    pool = mp.Pool(min(njobs, mp.cpu_count()))
    integrals_per_time = pool.starmap(___IntegralOfACF_chunk___, args_per_time)
    pool.close()
    integrals_per_time = np.asarray(integrals_per_time)
    # Saving data
    np.save(os.path.join(windows_path, 'IACFs_per_time.npy'), integrals_per_time)


    return integrals_per_window, integrals_per_time


def ___IntegralOfACF_chunk___(data, integrator):
    """Part of the code of IntegralOfACF
    """
    # index = int(os.path.basename(window))
    integrals = []
    for j in range(1, data.shape[1]):
        integrals.append(integrator(data[:,j], data[:,0]))
    return integrals

import numpy as np
import matplotlib.pyplot as plt

def custom_subplots(coords:list,x_size:int = 6):
    """This will create a grid plot where axes[3n+1,0] will be used for
    the contour plot, axes[3n,0] for the integral along the "X" and
    axes[3n+1,1] for the integral along the "Y"

    Parameters
    ----------
    coords : list
        A list of name of coordinates (could be numbers). The string Hello will be interpreted as 5 coordinates ['H', 'e', ...]
        The proper way is ['Hello']
    x_size : int, optional
        the size of the figure along "X", by default 6

    Returns
    -------
    tuples
        (fig, axes)
    """
    x_size = 6
    y_size = (len(coords))*x_size

    fig = plt.figure(figsize=(x_size, y_size)) # , layout="constrained"
    # Add a gridspec with two rows and two columns and a ratio of 1 to 4 between
    # the size of the marginal axes and the main axes in both directions.
    # Also adjust the subplot parameters for a square plot.

    gs = fig.add_gridspec(3*len(coords), 2, width_ratios=(4, 1), height_ratios=len(coords)*[1, 4, 0.5], # 1, 4, 0.5
                        left=0.1, right=0.9, bottom=0.1, top=0.9,
                        wspace=0.05, hspace=0.05)
    # Create the Axes.
    axes = np.empty([3*len(coords), 2], dtype=object)
    for i in range(len(coords)):
        axes[3*i+1,0] = fig.add_subplot(gs[3*i+1,0])
        axes[3*i,0] = fig.add_subplot(gs[3*i,0])#, sharex=axes[2*i+1,0])
        axes[3*i,0].set_xticklabels([])
        axes[3*i+1,1] = fig.add_subplot(gs[3*i+1,1])#, sharey=axes[2*i+1,0])
        axes[3*i+1,1].set_yticklabels([])

        axes[3*i,0].set(
            title = f'Coord {coords[i]}'
        )
    return (fig, axes)

fig, axes = custom_subplots('abc')


def plot(windows_path, ac):
    pattern = re.compile("^\d{5}$")
    individual_paths = sorted([os.path.join(windows_path, window) for window in tools.list_if_dir(windows_path) if pattern.match(window)])

    all_data = []
    for window in individual_paths:
        multidata = xvg.XVG(os.path.join(window, ac)).multidata

        data_on_window = multidata[0]
        for data in multidata[1:]:
            data_on_window = np.insert(data_on_window,-1, data[:,1], axis = -1)
        all_data.append(data_on_window)
    #all data shape = (number of windows, number of time points, number or reaction coordinates +1 [time is the first column])
    all_data = np.stack(all_data, axis = 0)
    # Getting the levels
    minimum = -0.5 # all_data[:,:,1:].min() # -1
    maximum = 1 # all_data[:,:,1:].max() # 1
    levels = np.linspace(minimum, maximum, 10)



    cm = plt.get_cmap('viridis')#gist_rainbow viridis

    fig, axes = plt.subplots(nrows =  all_data.shape[2]-1, ncols = 1, figsize = (5,10), constrained_layout=True)
    plt.setp(axes.flat, ylabel="Window ID" , xlabel= "Time [ps]")



    # for ax, col in zip(axes[0], ['System']):
    #     ax.annotate(col, xy=(0.5, 1), xytext=(0, pad),
    #                 xycoords='axes fraction', textcoords='offset points',
    #                 size='large', ha='center', va='baseline', fontsize = 30)

    # for ax, row in zip(axes[:,0], range(all_data.shape[0])):
    #     ax.annotate(str(row), xy=(0, 0.5), xytext=(-ax.yaxis.labelpad - pad, 0),
    #                 xycoords=ax.yaxis.label, textcoords='offset points',
    #                 size='large', ha='right', va='center', fontsize = 30)

    X, Y = np.meshgrid(range(1, all_data.shape[0]+1),all_data[0,:,0])
    X, Y = np.meshgrid(all_data[0,:,0], range(1, all_data.shape[0]+1))

    for i in range(1, all_data.shape[-1]):
        axes[i-1].set(
            title = f'coord {i}'
        )
        Z = np.reshape(all_data[:,:,i], X.shape)
        CS = axes[i-1].contourf(X, Y, Z, cmap = cm, levels = levels)

    cbar = fig.colorbar(CS, ax=axes)
    cbar.ax.get_yaxis().labelpad = 15
    cbar.ax.set_ylabel('ACF', rotation=270, fontsize = 15)


    fig.savefig("test.png", bbox_inches="tight")

def main(windows_path, pullx, ac = 'autocorr.xvg', overwrite = False, reaction_coordinates = None,  integrator:str = 'trapezoid', njobs = 1, **kwargs):

    pattern = re.compile("^\d{5}$")
    individual_paths = sorted([os.path.join(windows_path, window) for window in tools.list_if_dir(windows_path) if pattern.match(window)])

    ListOfKwargs = []
    extra_kwargs = kwargs.copy()
    for window in individual_paths:
        main_kwargs ={
            'f':os.path.join(window, pullx),
            'ac':os.path.join(window, ac),
            'overwrite':overwrite,
            'reaction_coordinates': reaction_coordinates,
            }
        if kwargs:
            extra_kwargs = kwargs.copy()
            extra_kwargs.update(main_kwargs)
            ListOfKwargs.append(extra_kwargs)
        else:
            ListOfKwargs.append(main_kwargs)

    print('Calculating ACFs..')
    pool = mp.Pool(min(njobs, mp.cpu_count()))
    pool.map(___gmx_analyze_wrapper___, ListOfKwargs)
    # pool.imap(gmx_analyze_wrapper, ListOfKwargs)
    pool.close()
    print('Done!')

    print('Calculating integral of ACFs per windows...')
    IntegralOfACF(windows_path = windows_path,  ac = ac, integrator =integrator, njobs=njobs)
    print('Done!')









if __name__ == '__main__':
    windows_path1 = '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_O/7e27/sys_MMV007839_Cell_891_SP_param/windows'
    windows_path2 = '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_Q/7e27/sys_MMV007839_Cell_891_SP_param/windows'
    windows_path3 = '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_A/7e27/sys_MMV007839_Cell_891_SP_param/windows'

    ww = [windows_path3]
    # main(
    #     windows_path = windows_path,
    #     pullx='production_pullx.xvg',
    #     ac = 'autocorr.xvg',
    #     overwrite=True,
    #     reaction_coordinates=[1, 2, 3, 4, 5],
    #     njobs=12
    # )
    # my_dict = {'f': '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_A/7e27/sys_MMV007839_Cell_891_SP_param/windows/00020/production_pullx.xvg', 'ac': '/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_A/7e27/sys_MMV007839_Cell_891_SP_param/windows/00020/autocorr.xvg', 'overwrite': False}
    # # gmx_analyze_wrapper(my_dict)
    # autocorr_xvg = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_A/7e27/sys_MMV007839_Cell_891_SP_param/windows/00020/autocorr.xvg')
    # print(autocorr_xvg.multidata[0])
    # print(len(autocorr_xvg.multidata))

    for i, w in enumerate(ww):
        a, b = IntegralOfACF(windows_path = w, njobs=12)
        print(a.shape, b.shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os, re, tempfile, argparse
from palmiche.utils import tools
import multiprocessing as mp
import tqdm



def main(input_path, conf_file_name, output_path = '.', cpu = 0):
    """Take the last frame of the production of each window, and then concatenate them.

    Args:
        input_path (path): The path were the windows directories are with name xxxxx, where x is a number
        conf_file_name (string): the name of the last frame exported when the umbrella is finished
        output_path (str, optional): The output path where the file cat_windows.xtc will be output. Defaults to '.'.
    """
    pattern = re.compile("^\d{5}$")
    conf_file_paths = sorted([os.path.join(input_path, window, conf_file_name) for window in tools.list_if_dir(input_path) if pattern.match(window)])
    temp_folder = tempfile.TemporaryDirectory(dir = '.', prefix='.')
    #with tempfile.TemporaryDirectory() as temp_folder:
    xtcs = []
    cmds = []
    for (i, conf) in enumerate(conf_file_paths):
        xtc_tmp = os.path.join(temp_folder.name, f"{i}.xtc")

        cmds.append(f"export GMX_MAXBACKUP=-1;  gmx trjconv -f {conf} -o {xtc_tmp}")
        # I catch the xtc names and then they will be ordered
        xtcs.append(xtc_tmp)
   #In order to parallelize the dist calculation
    if cpu:
        jobs = cpu
    else:
        jobs = mp.cpu_count()
    pool = mp.Pool(jobs)
    print("Converting the last frame of each window to .xtc...")
    for i in tqdm.tqdm(pool.imap_unordered(tools.run, cmds), total=len(cmds)):
        pass
    #pool.map(tools.run,cmds)
    pool.close()
    print("Done!")

    print("Concatening .xtc files...")
    tools.run(f"gmx trjcat -f {' '.join([xtc for xtc in xtcs])} -o {os.path.join(output_path, 'cat_windows.xtc')} -cat")
    print("Finished!")


if __name__ == '__main__':

    parser = argparse.ArgumentParser(description=__doc__,
                                     formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument('--input_path','-i',
                        help = "The path were the windows directories are with name xxxxx, where x is a number",
                        dest='input_path',
                        type=str)
    parser.add_argument('--conf_file_name', '-c',
                        help = "the name of the last frame exported when the umbrella is finished",
                        dest='conf_file_name',
                        type=str)
    parser.add_argument('--output_path', '-o',
                        help = "The output path where the file cat_windows.xtc will be output. Defaults to '.'",
                        default='.',
                        type=str)

    args = parser.parse_args()

    main(args.input_path, args.conf_file_name, args.output_path)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from collections import Counter

class MARKOV:
    def __init__(self, realization):
        self.realization = realization
        #>> matrix = Markov(realization)
        #>> matrix.compute_transition_matrix() will update the matrix if a different realization is passed
        self.compute_transition_matrix()
    
    def compute_transition_matrix(self):
        #Getting the states check what method to use for sorted the states
        if any(isinstance(n, str) for n in self.realization):
            states = sorted(set(self.realization), key=lambda x:str(x))
        else:
            states = sorted(set(self.realization))

        #Translate states to numerical indexes
        transdict = dict(zip(states, range(len(states))))

        # Converting the realization to numbers
        realization_to_work = [transdict[item] for item in self.realization]
        
        # Creating the matrix
        self.transition_matrix = np.zeros([len(states)]*2)
        for (i,j) in zip(realization_to_work,realization_to_work[1:]):
            self.transition_matrix [i][j] += 1
        
        # Convert to probability
        for row in self.transition_matrix :
            s = row.sum()
            if s > 0:
                row /= s

        # Adding labels
        self.transition_matrix = pd.DataFrame(self.transition_matrix, columns=states, index = states)
    
    def plot(self, out = 'transition_matrix.svg', occupancy = True):
        sns.set_theme(style="whitegrid")
        if occupancy:
            fig, ax = plt.subplots(nrows=1, ncols=2, figsize = (25,10))
            sns.heatmap(self.transition_matrix, ax = ax[0], vmin=.0, vmax=.5, center=.3, annot=True, cbar_kws={'label': 'Transition Probability'})
            df = pd.DataFrame(Counter(self.realization), index = [0])
            df /= df.values.sum()
            # Rearrange the columns
            cols = df.columns.tolist()
            if any(isinstance(n, str) for n in cols):
                cols = sorted(cols, key=lambda x:str(x))
            else:
                cols = sorted(cols)
            sns.barplot(data = df[cols], ax = ax[1])
            ax[0].set(
                xlabel = 'States',
                ylabel = 'States')
            ax[1].set(
                xlabel = 'States',
                ylabel = 'Probability')
        else:
            fig, ax = plt.subplots(figsize = (12,10))
            sns.heatmap(self.transition_matrix, ax = ax, vmin=.0, vmax=.5, center=.3, annot=True, cbar_kws={'label': 'Transition Probability'})
            ax.set(
                xlabel = 'States',
                ylabel = 'States'
                )
        fig.savefig(out, bbox_inches="tight")
        sns.reset_defaults()

        
# def transition_matrix(realization):
#     #The convertion is in case different types of data are passed
#     realization = np.array(realization)
    
#     #Getting the states
#     states = np.sort(np.unique(realization))

#     #Translate states to numerical indexes
#     transdict = dict(zip(states, range(len(states))))

#     # Converting the realization to numbers
#     realization_to_work = [transdict[item] for item in realization]
    
#     # Creating the matrix
#     M = np.zeros([states.shape[0]]*2)
#     for (i,j) in zip(realization_to_work,realization_to_work[1:]):
#         M[i][j] += 1
    
#     # Convert to probability
#     for row in M:
#         s = row.sum()
#         if s > 0:
#             row /= s
    
#     # Adding labels
#     M = pd.DataFrame(M, columns=states, index = states)
#     return M

if __name__ == '__main__':
    import matplotlib.pyplot as plt
    import seaborn as sns
    from palmiche.utils import xvg
    realization = xvg.XVG('/home/ale/mnt/smaug/MD/NEW/docking_min_equi/umbrella_iteration/umbrella_P/7e27/sys_MMV007839_Cell_891_SP_param/windows/00001/production.xvg').data[:,1]
    # realization = [1,1,1,2,2,'B',4,4,8,8,55,47,'A', 'B', 'A']
    # realization = [1,1,1,2,2,4,4,8,8,55,47,5,5,1,1,2,3,10,12,13,13,13,13,15,15,45,47,48,47,45,46,49,48,47]
    plt.plot(realization)
    plt.show()
    A = MARKOV(realization.astype(int))
    A.plot('p.pdf')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from palmiche.utils import tools, mdp, xvg, markov
from palmiche.simultemp import weights
import os
import numpy as np
import matplotlib.pyplot as plt
import socket


def main(wd = './',
        weights_method = 'PandeWeights', #'AleWeights', 'PandeWeights' esthetic
        EnergyType = 'Potential', # 'Total-Energy'
        heat_fraction = 0.25, # DEpendding on what weights_method is chosen heat_fraction or heating_time and constant_temp_time will be used
        heating_time = 100,
        constant_temp_time = 2000,
        number_of_tc_grps = 1,
        temp_list = range(300, 332, 2),
        maxsol = 500,
        box = (2.5,2.5,2.5),
        water_ff = 'tip3p',
        ions_ff = 'oplsaa',
        simulation_time = {'annealing':60000,'tempering': 60000}, # ps the annealing time only will be used for the method AleWeights, and is a similar time to the one use for the calculation of PandeWeights
        xtc_numb_frame = 1000,
        weight_GROMACS_format = True,
        cpu = 12,
        # This arguments will give the option to only collect those point in an specific 
        # interval of time. E. g = (500, 670) will collect those point between 500 and 670 ps,
        # None means that all point will be collected. This feature could be useful in case that
        # only some region of the simulation want to be analyzed (the coling region, for example).
        # the interval is closed, both extrames are taken
        collecting_interval = None,
        SIMTEMP_OCCUP_STATE0 = 0,
        **mdp_options,):

    if SIMTEMP_OCCUP_STATE0:
       str_SIMTEMP_OCCUP_STATE0 = f"""
       source /data/shared/opt/gromacs/2021.6_simTemp_setOccupState0/bin/GMXRC.bash
       PATH=$PATH:"/data/shared/opt/gromacs/2021.6_simTemp_setOccupState0/bin"
       export SIMTEMP_OCCUP_STATE0={SIMTEMP_OCCUP_STATE0}
       """
    else:
        str_SIMTEMP_OCCUP_STATE0 = '#' 

    tools.makedirs(wd)
    hostname = socket.gethostname()
    if hostname == 'smaug':
        mdrun_opt = '-pin on -pinstride 1 -pinoffset 0 -gpu_id 0 -nice 19'
    else:
        mdrun_opt = '-pin on -pinstride 1 -pinoffset 0 -nice 19'

    minimization_MDP =mdp.MDP(type = 'minimization',**mdp_options)
    minimization_MDP.write(os.path.join(wd,'minimization.mdp'))

    annealing_MDP = mdp.MDP(
        type = 'production',
        time = simulation_time['annealing'],
        xtc_numb_frame = xtc_numb_frame,
        **mdp_options)
    # Choose what kind of annealing strategy depending the weight calculation method
    if weights_method == 'PandeWeights':
        annealing_temp, annealing_time = annealing_MDP.annealing3(number_of_tc_grps,temp_list, heating_time = heating_time, constant_temp_time = constant_temp_time)
    elif weights_method == 'AleWeights':
        annealing_MDP.annealing1(number_of_tc_grps,temp_list,heat_fraction=heat_fraction,  nstenergy_same_as_nstcalcenergy = True)
    annealing_MDP.write(os.path.join(wd,'annealing.mdp'))

    cwd = os.getcwd()
    os.chdir(wd)

    tools.run(f"""
        {str_SIMTEMP_OCCUP_STATE0}
        export GMX_MAXBACKUP=-1
        gmx solvate -box {' '.join([str(b) for b in box])} -maxsol {maxsol} -o water.gro
        gmx pdb2gmx -f water.gro -o water.gro -water {water_ff} -ff {ions_ff} -p topol.top
        echo q | gmx make_ndx -f water.gro -o index.ndx

        # This is for testing, it should be at the job.sh, and them add the wait function of tools module
        # Or even better launch the script itself. And them everything is done nicely. No need to wait an all of that
        gmx grompp -f minimization.mdp  -c water.gro -p topol.top -n index.ndx -o minimization.tpr
        gmx mdrun -nt {cpu} -stepout 5000 -v -deffnm minimization -cpi -maxh 48 {mdrun_opt} >& minimization.lis
        gmx grompp -f annealing.mdp  -c minimization.gro -p topol.top -n index.ndx -o annealing.tpr
        gmx mdrun -nt {cpu} -stepout 5000 -v -deffnm annealing -cpi -maxh 48 {mdrun_opt} >& annealing.lis
        """)
    
    if weights_method == 'PandeWeights':
        energy = weights.get_energy('annealing.edr', annealing_time, energy_type = EnergyType)

        W = weights.PandeWeights(
                temperatures = sorted(set(annealing_temp)),
                energies = energy,
                alpha = SIMTEMP_OCCUP_STATE0, 
                weight_GROMACS_format = weight_GROMACS_format
                )

    elif weights_method == 'AleWeights':

        """
        Getting the function E(T), the way that will be done is averaging the internal energy 
        The same value of temperature could have several values of total energy
        The first an easiest approximation is take all the simulation into account
        The second one is just take into account the cooling region
        """

        tools.run(f"""
        {str_SIMTEMP_OCCUP_STATE0}
        export GMX_MAXBACKUP=-1
        echo Temperature | gmx energy -f annealing.edr -s annealing.tpr -o temperature.xvg
        echo {EnergyType} | gmx energy -f annealing.edr -s annealing.tpr -o energy.xvg
        """)
        temperature = xvg.XVG('temperature.xvg').data
        energy = xvg.XVG('energy.xvg').data

        # In case that collecting_interval is provided
        if collecting_interval:
            collecting_interval = np.array(collecting_interval)*1000
            # Could also been selected energy because the first column (time) is the same
            mask = (temperature[:,0] >= collecting_interval.min()) & (temperature[:,0] <= collecting_interval.max())
            temperature = temperature[mask]
            energy = energy[mask]


        # df = pd.DataFrame(np.vstack([temperature[:,1], energy[:,1]]).T, columns = ['temperature', 'energy'])
        # mean = df.groupby(['temperature']).mean()
        # mean.sort_index(inplace=True)
        # # # To get a column named temperature
        # # mean['temperature'] = mean.index
        # # mean.reset_index(drop=True, inplace=True)
    
        # Only take into account the study interval of temperature
        # The "-2" rectification is to take into account a little more at the edges, but it will not affect the calculation of the weights
        # It will only add extra points. The problem is that with the filter let say that we add all the number greater than or equal to 300 K
        # But then in the integration 300 is not in the DataFrame, is just for precaution.
        
        mask = (temperature[:,1] >= min(temp_list) - 2) & (temperature[:,1] <= max(temp_list) + 2)
        temperature = temperature[mask]
        energy = energy[mask]

        fig, ax = plt.subplots(figsize = (25,9))
        ax.scatter(temperature[:,1], energy[:,1])
        ax.set(
            xlim = [min(temp_list), max(temp_list)],
            xlabel = 'T [K]',
            ylabel = f'{EnergyType} [kJ/mol]',
            title = f'Dependency of the {EnergyType} with the Temperature'
        )
        fig.savefig(f'{EnergyType}_vs_T.png')
        
        # Calculating the weights
        W = weights.AleWeights(
            temperature[:,1],
            energy[:,1],
            temp_list,
            alpha = SIMTEMP_OCCUP_STATE0,
            weight_GROMACS_format = weight_GROMACS_format)

    # Simulated Tempering
    tempering_MDP = mdp.MDP(
        type = 'production',
        time = simulation_time['tempering'],
        xtc_numb_frame = xtc_numb_frame,
        **mdp_options)
    tempering_MDP.simulated_tempering(temp_list, W)
    tempering_MDP.write(os.path.join('tempering.mdp'))

    tools.run(f"""
        {str_SIMTEMP_OCCUP_STATE0}
        gmx grompp -f tempering.mdp  -c minimization.gro -p topol.top -n index.ndx -o tempering.tpr
        gmx mdrun -nt {cpu} -stepout 5000 -v -deffnm tempering -cpi -maxh 48 {mdrun_opt} >& tempering.lis
        """)
    weights.get_weights_from_log('tempering.log', plot = True)
    
    tempering_xvg_data = xvg.XVG('tempering.xvg').data
    markov.MARKOV(tempering_xvg_data[:,1].astype(int)).plot('transition_matrix.svg')


    os.chdir(cwd)

if __name__ == '__main__':
    #ant the value to 20000 to get more resolution on the energy on this step
    # I need to improve this part because n the simulated tempering I dont need such resolution on the energy
    tc_grps = ['Water']
    ref_t = 300
    # I put const
    mdp_options = {             
    "tc_grps":' '.join(tc_grps),
    "tau_t":' '.join(len(tc_grps)*['1.0']),
    "ref_t":' '.join(len(tc_grps)*[str(ref_t)]),
    "dt": 0.002,
    "pcoupltype": 'isotropic',
    "compressibility": 4.5e-5,
    "ref_p": 1.0,
    #'rlist': 1,
    #'rcoulomb': 1,
    }
    args = [
            {
                    'wd' : 'PandeWeights',
                    'weights_method' : 'PandeWeights',
                    'EnergyType':'Potential', # 'Total-Energy'
                    'heat_fraction' : 0.25,
                    'heating_time' : 2,
                    'constant_temp_time' : 5,
                    'number_of_tc_grps' : 1,
                    'temp_list' : range(300, 332, 2),
                    'maxsol' : 500,
                    'box' : (2.5,2.5,2.5),
                    'water_ff' : 'tip3p',
                    'ions_ff' : 'oplsaa',
                    'simulation_time' : {'annealing':300,'tempering': 50}, #ps
                    'xtc_numb_frame' : 1000,
                    'weight_GROMACS_format' : True,
                    'cpu' : 12,
                    'collecting_interval' : None,
                    'SIMTEMP_OCCUP_STATE0': 0,
            },
            {
                    'wd' : 'AleWeights',
                    'weights_method' : 'AleWeights',
                    'EnergyType':'Potential', # 'Total-Energy'
                    'heat_fraction' : 0.25,
                    'heating_time' : 2,
                    'constant_temp_time' : 5,
                    'number_of_tc_grps' : 1,
                    'temp_list' : range(300, 332, 2),
                    'maxsol' : 500,
                    'box' : (2.5,2.5,2.5),
                    'water_ff' : 'tip3p',
                    'ions_ff' : 'oplsaa',
                    'simulation_time' : {'annealing':300,'tempering': 50}, #ps
                    'xtc_numb_frame' : 1000,
                    'weight_GROMACS_format' : True,
                    'cpu' : 12,
                    'collecting_interval' : None,    
                    'SIMTEMP_OCCUP_STATE0':0, 
            }
    ]
    for arg in args:
        arg.update(mdp_options)
        main(**arg)